"""
Brython MDCComponent: MDCMenu
=============================


"""


from browser import window
from .core import MDCTemplate
# from .MDCButton import MDCButton,  MDCIconToggle


########################################################################
class __menuItem__(MDCTemplate):
    """"""

    MDC_optionals = {

        # 'disable': 'tabindex="-1" aria-disabled="true"',
        # 'stack' = '{}'
        'icon': '<i class="radiant-menu-icon material-icons" aria-hidden="true" style="margin-right: 15px;">{icon}</i>',
        'fa_icon': '<i class="radiant-menu-icon {fa_style} {fa_icon}" style="margin-right: 15px;"></i>',

    }

    # ----------------------------------------------------------------------
    def __new__(self, text, icon=None, stack_icon='', disable=False, **kwargs):
        """"""

        if icon and icon.startswith('fa'):
            fa_style = icon[:icon.find('-')]
            fa_icon = 'fa' + icon[icon.find('-'):]
            fa_icon_ = fa_icon
            del icon

        self.element = self.render(locals(), kwargs)

        return self.element

    # ----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""

        if not context['disable']:
            cls.MDC_optionals['disable'] = 'tabindex="0"'
        else:
            cls.MDC_optionals['disable'] = 'tabindex="-1" aria-disabled="true"',

        code = """
            <li class="mdc-list-item" role="menuitem" {disable}>
                {stack_icon}
                {icon}
                {fa_icon}
                <span class="mdc-list-item__text">{text}</span>
            </li>

        """
        return cls.render_html(code, context)

    # # ----------------------------------------------------------------------
    # @classmethod
    # def __getitem__(self, name):
        # """"""
        # if name is 'icon':
            # return self.element.select('.radiant-menu-icon')[0]


########################################################################
class MDCMenu(MDCTemplate):
    """"""

    NAME = 'menu', 'MDCMenu'
    # NAME = 'menuSurface', 'MDCMenuSurface'

    CSS_classes = {

        # '_16_9': 'mdc-card__media--16-9',
        # 'square': 'mdc-card__media--square',
    }

    MDC_optionals = {

        # 'outlined': 'mdc-card--outlined',
        # 'full_bleed': 'mdc-card__actions--full-bleed',
        # 'icon': '<i class="material-icons mdc-button__icon" aria-hidden="true">{icon}</i>',
        # 'disabled': 'disabled',

    }

    # ----------------------------------------------------------------------
    def __new__(self, corner='BOTTOM_END', autoclose=True, **kwargs):
        """"""

        self.element = self.render(locals(), kwargs)

        # if corner in ['BOTTOM_START', 'BOTTOM_LEFT', 'BOTTOM_RIGHT', 'BOTTOM_END', 'TOP_START', 'TOP_LEFT', 'TOP_RIGHT', 'TOP_END']:
            # c = getattr(window.mdc.menu.MDCMenuFoundation.Corner, corner)
            # print(c, self.mdc.setAnchorCorner)
        # self.mdc.setAnchorCorner(window.mdc.menu.MDCMenuFoundation.Corner.BOTTOM_END)

        if autoclose:
            self.element.bind('click', self.toggle)

        return self.element


# menu.setAnchorCorner(mdc.menu.MDCMenuFoundation.Corner.TOP_START)

    # ----------------------------------------------------------------------


    @classmethod
    def __html__(cls, **context):
        """"""

        code = """

        <div class="mdc-menu mdc-menu-surface" tabindex="-1">
          <ul class="mdc-list" role="menu" aria-hidden="true" aria-orientation="vertical">

          </ul>
        </div>

        """
        return cls.render_html(code, context)

    # ----------------------------------------------------------------------

    @classmethod
    def __getitem__(self, name):
        """"""
        if name is 'content':
            return self.element.select('.mdc-list')[0]

        # elif name is 'action_buttons':
            # return self.element.select('.mdc-card__action-buttons')[0]

        # elif name is 'action_icons':
            # return self.element.select('.mdc-card__action-icons')[0]

    # ----------------------------------------------------------------------

    @classmethod
    def add_item(cls, element, *args, **kwargs):
        """"""
        item = __menuItem__(*args, **kwargs)
        cls['content'] <= item
        return item

    # ----------------------------------------------------------------------

    @classmethod
    def add_divider(cls, element):
        """"""
        divider = '<li class="mdc-list-divider" role="separator"></li>'
        divider = cls.render_str(divider)
        cls['content'] <= divider

    # ----------------------------------------------------------------------

    @classmethod
    def toggle(cls, element, corner='TOP_START', *args, **kwargs):
        """"""
        if corner in ['BOTTOM_START', 'BOTTOM_LEFT', 'BOTTOM_RIGHT', 'BOTTOM_END', 'TOP_START', 'TOP_LEFT', 'TOP_RIGHT', 'TOP_END']:
            c = getattr(window.mdc.menuSurface.Corner, corner)
            cls.mdc.setAnchorCorner(c)

        cls.mdc.open = not cls.mdc.open
        # self['title'].text = text





"""
Brython MDCComponent: MDCShape
==============================


"""

from .core import MDCTemplate
#from .MDCButton import MDCButton,  MDCIconToggle


########################################################################
class MDCShape(MDCTemplate):
    """"""

    NAME = 'shape', 'MDCShape'

    CSS_classes = {

        #'_16_9': 'mdc-card__media--16-9',
        #'square': 'mdc-card__media--square',
    }

    MDC_optionals = {

        #'outlined': 'mdc-card--outlined',
        #'full_bleed': 'mdc-card__actions--full-bleed',
        #'icon': '<i class="material-icons mdc-button__icon" aria-hidden="true">{icon}</i>',
        #'disabled': 'disabled',

    }

    #----------------------------------------------------------------------
    def __new__(self, component, **kwargs):
        """"""
        #component = component
        self.element = self.render(locals(), kwargs)

        self.element <= component

        return self.element



    #----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""

        #context['component'] = context['component'].html

        #cls.ID = cls.new_id()
        #context['id'] = cls.ID

        code = """
            <div class="mdc-shape-container">
                <div class="mdc-shape-container__corner mdc-shape-angled-corner mdc-shape-container__corner--top-left"></div>
                <div class="mdc-shape-container__corner mdc-shape-container__corner--top-right"></div>
                <div class="mdc-shape-container__corner mdc-shape-container__corner--bottom-right"></div>
                <div class="mdc-shape-container__corner mdc-shape-container__corner--bottom-left"></div>
              </div>
        """



        return cls.render_html(code, context)


    #----------------------------------------------------------------------
    @classmethod
    def set_corners(cls, element, background_color='unset', outlined=False, size=None, outline_width='2px', outline_style='solid', outline_color='unset'):
        """"""

        if not isinstance(size, (list, tuple)):
            args = size, size, size, size,
        else:
            args = size

        if len(args) > 1:
            top_left_size, top_right_size, bottom_right_size, bottom_left_size = args
        else:
            top_left_size, top_right_size, bottom_right_size, bottom_left_size = args[0], args[0], args[0], args[0]

        corners = (
            ('top', 'left', top_left_size),
            ('top', 'right', top_right_size),
            ('bottom', 'right', bottom_right_size),
            ('bottom', 'left', bottom_left_size),
        )

        for x, y, size in corners:

            diagonal_length = 'calc({} * 1.4142135623730951)'.format(size)
            diagonal_length_2 = 'calc({} * -1.4142135623730951 / 2)'.format(size)

            corner = cls.element.select('.mdc-shape-container__corner--{}-{}'.format(x, y))[0]

            corner.style = {
                '{}'.format(y): diagonal_length_2,
                '{}'.format(x): diagonal_length_2,
                'width': diagonal_length,
                'height': diagonal_length,
                'background-color': background_color,
            }


        if outlined:

            for shape in cls.element.select('.mdc-shape-container__corner'):
                shape.style = {
                    #'top': outline_width,
                    'border-bottom': '{} {} {}'.format(outline_width, outline_style, outline_color),
                }


    #@mixin mdc-shape-angled-corner-outline($outline-width, $outline-color, $outline-style: solid) {
      #.mdc-shape-container__corner::before {
        #top: $outline-width;
        #border-bottom: $outline-width $outline-style $outline-color;
      #}
    #}




        #@each $y, $x, $size in $corners {
          #@if $size > 0 {
            #// Assume that the size passed represents the width/height rather than the diagonal.
            #// This makes it trivial e.g. to accomplish a horizontal arrow cap by specifying exactly half the shape's height.
            #// Unfortunately, Sass doesn't have built-in exponentiation, and node-sass custom functions seem to hang.
            #// (https://github.com/sass/node-sass/issues/857 may be related, but the workaround didn't seem to work.)
            #// Fortunately, if we assume 45 degree cutouts all the time, a^2 + b^2 = c^2 simplifies to a*sqrt(2).
            #$diagonal-length: $size * 1.4142135623730951;

            #.mdc-shape-container__corner--#{$y}-#{$x} {
              ##{$y}: -$diagonal-length / 2;
              ##{$x}: -$diagonal-length / 2;
              #width: $diagonal-length;
              #height: $diagonal-length;
            #}
          #}
        #}


    #----------------------------------------------------------------------
    @classmethod
    def __getitem__(self, name):
        """"""
        #if name is 'actions':
            #return self.element.select('.mdc-card__actions')[0]

        #elif name is 'action_buttons':
            #return self.element.select('.mdc-card__action-buttons')[0]

        #elif name is 'action_icons':
            #return self.element.select('.mdc-card__action-icons')[0]


    #----------------------------------------------------------------------
    @classmethod
    def add_action_button(cls, element, mdc, *args, **kwargs):
        """"""



    ##----------------------------------------------------------------------
    #@classmethod
    #def title(self, mdc, text):
        #""""""
        #self['title'].text = text





"""
============
EEG Features
============

"""

import numpy as np
from scipy.fftpack import fft, fftfreq, fftshift
from scipy.signal import welch as sci_welch
from datetime import datetime


# ----------------------------------------------------------------------
def entropy(eeg):
    """Signal entropy."""

    upper = np.mean(eeg, axis=1) + 3 * np.std(eeg, axis=1)
    lower = np.mean(eeg, axis=1) - 3 * np.std(eeg, axis=1)

    bins = np.linspace(lower, upper, 20, axis=1)

    counts = [np.histogram(*xbin)[0] for xbin in zip(eeg, bins)]

    p = counts / np.sum(counts, axis=1)[:, None]
    p = [np.compress(p_ != 0, p_) for p_ in p]  # remove zeros

    H = np.array([-np.sum(p_ * np.log(p_)) / len(p_) for p_ in p])
    return H


# ----------------------------------------------------------------------
# def energy(eeg, / , axis=-1):
def energy(eeg, axis=-1):
    """Signal energy."""

    E = np.sum(abs(eeg ** 2), axis=axis)
    return E


# ----------------------------------------------------------------------
# def std(eeg, / , axis=-1):
def std(eeg, axis=-1):
    """Standard deviation."""

    STD = np.std(eeg, axis=axis)
    return STD


# ----------------------------------------------------------------------
# def spectrum(eeg, /, d=None, fs=None, timestamp=None, axis=-1):
def spectrum(eeg, d=None, fs=None, timestamp=None, axis=-1):
    """Spectral power density using Fourier."""

    timestamp = np.array(timestamp)

    if timestamp.any():
        if isinstance(timestamp[-1], (int, float)):
            delta = datetime.fromtimestamp(timestamp[-1]) - datetime.fromtimestamp(timestamp[0])
        else:
            delta = timestamp[-1] - timestamp[0]
        fs = max(eeg.shape) / delta.total_seconds()

    if fs:
        d = 1 / fs

    if axis is None:
        if eeg.shape[0] > eeg.shape[1]:
            axis = 0
        else:
            axis = 1

    Yf = fftshift(np.abs(fft(eeg, axis=axis)))
    f = fftshift(fftfreq(Yf.shape[1], d))

    n = (Yf.shape[1] // 2)
    Yf = Yf[:, n:] * (2 / Yf.shape[1])
    f = f[n:]

    return f, Yf


# ----------------------------------------------------------------------
# def welch(eeg, /, d=None, fs=None, timestamp=None, axis=-1):
def welch(eeg, d=None, fs=None, timestamp=None, axis=-1):
    """Spectral power density using Welch's method."""

    timestamp = np.array(timestamp)

    if timestamp.any():
        if isinstance(timestamp[-1], (int, float)):
            delta = datetime.fromtimestamp(timestamp[-1]) - datetime.fromtimestamp(timestamp[0])
        else:
            delta = timestamp[-1] - timestamp[0]
        fs = max(eeg.shape) / delta.total_seconds()

    if fs:
        d = 1 / fs

    if axis is None:
        if eeg.shape[0] > eeg.shape[1]:
            axis = 0
        else:
            axis = 1

    f, p = sci_welch(eeg, 1 / d, window='flattop', nperseg=min(eeg.shape[1], 1024), scaling='spectrum', axis=axis)
    return f, p





"""
Brython MDCComponent: MDCLinearProgress
=======================================


"""


from .core import MDCTemplate
#from .MDCButton import MDCButton,  MDCIconToggle

########################################################################
class MDCLinearProgress(MDCTemplate):
    """"""

    NAME = 'linearProgress', 'MDCLinearProgress'

    CSS_classes = {

        #'_16_9': 'mdc-card__media--16-9',
        #'square': 'mdc-card__media--square',
    }

    MDC_optionals = {

        'indeterminate': 'mdc-linear-progress--indeterminate',
        'reversed': 'mdc-linear-progress--reversed',

        #'outlined': 'mdc-card--outlined',
        #'full_bleed': 'mdc-card__actions--full-bleed',
        #'icon': '<i class="material-icons mdc-button__icon" aria-hidden="true">{icon}</i>',
        #'disabled': 'disabled',

    }

    #----------------------------------------------------------------------
    def __new__(self, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)
        return self.element



    #----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""

        code = """
            <div role="progressbar" class="mdc-linear-progress {indeterminate} {reversed}">
              <div class="mdc-linear-progress__buffering-dots"></div>
              <div class="mdc-linear-progress__buffer"></div>
              <div class="mdc-linear-progress__bar mdc-linear-progress__primary-bar">
                <span class="mdc-linear-progress__bar-inner"></span>
              </div>
              <div class="mdc-linear-progress__bar mdc-linear-progress__secondary-bar">
                <span class="mdc-linear-progress__bar-inner"></span>
              </div>
            </div>
        """


        return cls.render_html(code, context)



    #----------------------------------------------------------------------
    @classmethod
    def set_progress(cls, element, val, buffer=None):
        """"""
        cls.mdc.progress = val
        if buffer is None:
            buffer = 1
        cls.mdc.buffer = buffer


    #----------------------------------------------------------------------
    @classmethod
    def set_reverse(cls, element, val):
        """"""
        cls.mdc.reverse = val


    #----------------------------------------------------------------------
    @classmethod
    def set_determinate(cls, element, val):
        """"""
        cls.mdc.determinate = val



    ##----------------------------------------------------------------------
    #@classmethod
    #def __getitem__(self, name):
        #""""""
        #if name is 'progress':
            #return self.element.progress

        #elif name is 'action_buttons':
            #return self.element.select('.mdc-card__action-buttons')[0]

        #elif name is 'action_icons':
            #return self.element.select('.mdc-card__action-icons')[0]


            ##----------------------------------------------------------------------
            #@classmethod
            #def progress(cls, mdc, *args):
                #""""""
                #mdc.progress = *args



    ##----------------------------------------------------------------------
    #@classmethod
    #def title(self, mdc, text):
        #""""""
        #self['title'].text = text





import netifaces
import requests
import nmap
from tornado.web import RequestHandler
from openbci.utils.scan_wifi_modules import scan


########################################################################
class HomeHandler(RequestHandler):
    """"""

    # -------------------------------------------------------------------
    def get(self):
        """"""
        self.render('html/home.html')


########################################################################
class EndPointHandler(RequestHandler):

    # ----------------------------------------------------------------------
    def get(self):
        """"""
        command = self.get_argument('command')
        response = getattr(self, f"command_{command}")()
        self.write(response)

    # ----------------------------------------------------------------------
    def command_scan(self):
        """"""
        boards = {}

        for interface in netifaces.interfaces():

            if not (interface.startswith("wlan") or interface.startswith("wlp")):
                continue

            local_lan = netifaces.ifaddresses(interface)[netifaces.AF_INET][0]['addr']
            nm = nmap.PortScanner()
            nm.scan(hosts=f'{local_lan}/24', arguments='-sn')
            hosts = nm.all_hosts()

            for host in hosts:
                try:
                    response = requests.get(f'http://{host}/board', timeout=0.1)
                    if response.ok:
                        boards[host] = response.json()
                except:
                    continue

        return boards





"""
Brython MDCComponent: MDCFormField
==================================


"""

from .core import MDCTemplate
from browser import html

########################################################################


class MDCFormField(MDCTemplate):
    """"""
    NAME = 'formField', 'MDCFormField'

    MDC_optionals = {
        'end': 'mdc-form-field--align-end',
    }

    # ----------------------------------------------------------------------

    def __new__(self, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)

        self.element.style = {'width': '100%',
                              'min-height': '48px',
                              # 'margin-bottom': '8px',
                              }

        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
            <div class="mdc-form-field {end}"></div>
        """
        return cls.render_html(code, context)


########################################################################
class MDCForm(MDCTemplate):
    """"""

    # ----------------------------------------------------------------------
    def __new__(self, separator=None, formfield=None, formfield_style={}, **kwargs):
        """"""
        self.separator = separator
        self.formfield = formfield
        self.formfield_style = formfield_style
        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
            <form></form>
        """
        return cls.render_html(code, context)

    # ----------------------------------------------------------------------

    @classmethod
    def __genericElement__(cls, parent, Element, extern_label, *args, **kwargs):
        """"""
        if cls.formfield:
            formfield = cls.formfield.clone()
        else:
            formfield = MDCFormField(style=cls.formfield_style)

        if cls.formfield_style:
            formfield = MDCFormField(style=cls.formfield_style)

        if 'formfield' in kwargs:
            formfield = kwargs['formfield']

        el = Element(*args, **kwargs)
        formfield <= el
        if extern_label:
            formfield <= __formLabel__(id_=el.mdc.get_id(), label=args[0])
        if kwargs.get('helper_text', False):
            # print(kwargs.get('helper_text_persistent'))
            formfield <= __formHelper__(id_=el.mdc.get_id(), label=kwargs.get('helper_text'), persistent=kwargs.get('helper_text_persistent', True))
        # cls.element <= formfield
        parent <= formfield

        if cls.separator:
            # cls.element <= cls.separator.clone()
            parent <= cls.separator.clone()

        return el, formfield

    # ----------------------------------------------------------------------

    @classmethod
    def Checkbox(cls, element, *args, **kwargs):
        """"""
        el, element = cls.__genericElement__(element, MDCCheckbox, True, *args, **kwargs)
        return el

    # ----------------------------------------------------------------------

    @classmethod
    def Radio(cls, element, *args, **kwargs):
        """"""
        el, element = cls.__genericElement__(element, MDCRadio, True, *args, **kwargs)
        return el

    # ----------------------------------------------------------------------

    @classmethod
    def Select(cls, element, *args, **kwargs):
        """"""
        el, element = cls.__genericElement__(element, MDCSelect, False, *args, **kwargs)
        return el

    # ----------------------------------------------------------------------

    @classmethod
    def Slider(cls, element, *args, **kwargs):
        """"""
        el, element = cls.__genericElement__(element, MDCSlider, False, *args, **kwargs)
        return el

    # ----------------------------------------------------------------------

    @classmethod
    def Switch(cls, element, *args, **kwargs):
        """"""
        el, element = cls.__genericElement__(element, MDCSwitch, True, *args, **kwargs)
        return el

    # ----------------------------------------------------------------------

    @classmethod
    def TextField(cls, element, *args, **kwargs):
        """"""
        el, element = cls.__genericElement__(element, MDCTextField, False, *args, **kwargs)
        element.style = {'display': 'flow-root'}
        el.style = {'margin-top': 'unset'}
        return el

    # ----------------------------------------------------------------------

    @classmethod
    def TextAreaField(cls, element, *args, **kwargs):
        """"""
        el, element = cls.__genericElement__(element, MDCTextAreaField, False, *args, **kwargs)
        element.style = {'display': 'flow-root'}
        el.style = {'margin-top': '8px'}
        return el


########################################################################
class __formLabel__(MDCTemplate):
    """"""

    # ----------------------------------------------------------------------
    def __new__(self, id_, label):
        """"""
        code = """
            <label for="{id_}">{label}</label>
            """.format(id_=id_, label=label)
        return self.render_str(code)


########################################################################
class __formHelper__(MDCTemplate):
    """"""
    NAME = 'textField', 'MDCTextFieldHelperText'

    # ----------------------------------------------------------------------
    def __new__(self, id_, label, persistent=True):
        """"""
        if persistent:
            persistent = 'mdc-text-field-helper-text--persistent'
        else:
            persistent = ''
        code = """
            <p id="{id_}" class="mdc-text-field-helper-text {persistent}" aria-hidden="true">
              {label}
            </p>
            """.format(id_=id_, label=label, persistent=persistent)
        return self.render_str(code)


########################################################################
class MDCCheckbox(MDCTemplate):
    """"""
    NAME = 'checkbox', 'MDCCheckbox'

    MDC_optionals = {

        'disabled': 'mdc-checkbox--disabled',
        'checked': 'checked',

    }

    # ----------------------------------------------------------------------

    def __new__(self, label, name='', value='', checked=False, disabled=False, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)
        # self.element.need_label = True
        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""

        cls.ID = cls.new_id()
        context['id'] = cls.ID

        if context['disabled']:
            context['input_disabled'] = 'disabled'
        else:
            context['input_disabled'] = ''

        code = """
              <div class="mdc-checkbox {disabled}">
                <input type="checkbox"
                       id="{id}"
                        name="{name}"
                        value="{value}"
                       class="mdc-checkbox__native-control"
                       {input_disabled}
                       {checked}
                       />
                <div class="mdc-checkbox__background">
                  <svg class="mdc-checkbox__checkmark"
                       viewBox="0 0 24 24">
                    <path class="mdc-checkbox__checkmark-path"
                          fill="none"
                          stroke="white"
                          d="M1.73,12.91 8.1,19.28 22.79,4.59"/>
                  </svg>
                  <div class="mdc-checkbox__mixedmark"></div>
                </div>
              </div>
        """
        return cls.render_html(code, context)


########################################################################
class MDCRadio(MDCTemplate):
    """"""
    NAME = 'radio', 'MDCRadio'

    MDC_optionals = {

        'disabled': 'mdc-radio--disabled',
        'checked': 'checked',

    }

    # ----------------------------------------------------------------------

    def __new__(self, label, name='', checked=False, disabled=False, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)
        # self.element.need_label = True
        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""

        cls.ID = cls.new_id()
        context['id'] = cls.ID

        if context['disabled']:
            context['input_disabled'] = 'disabled'
        else:
            context['input_disabled'] = ''

        code = """
            <div class="mdc-radio {disabled}">
              <input class="mdc-radio__native-control" type="radio" id="{id}" name="{name}" {input_disabled} {checked}>
              <div class="mdc-radio__background">
                <div class="mdc-radio__outer-circle"></div>
                <div class="mdc-radio__inner-circle"></div>
              </div>
            </div>
        """
        return cls.render_html(code, context)


########################################################################
class MDCSelect(MDCTemplate):
    """"""
    NAME = 'select', 'MDCSelect'

    MDC_optionals = {

        'disabled': 'mdc-select--disabled',
        'box': 'mdc-select--box',

    }

    # ----------------------------------------------------------------------

    def __new__(self, label, options=[], selected=None, disabled=False, box=False, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)
        # self.element.need_label = False

        if options:
            self.add_options(self.element, options, selected)

        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""

        cls.ID = cls.new_id()
        context['id'] = cls.ID

        if context['disabled']:
            context['input_disabled'] = 'disabled'
        else:
            context['input_disabled'] = ''

        code = """
            <div class="mdc-select {box} {disabled}" style="width: 100%">
              <select class="mdc-select__native-control" {input_disabled}>

              </select>
              <label class="mdc-floating-label">{label}</label>
              <div class="mdc-line-ripple"></div>
            </div>
        """
        return cls.render_html(code, context)

    # ----------------------------------------------------------------------

    @classmethod
    def __getitem__(self, name):
        """"""
        if name is 'options_placeholder':
            return self.element.select('.mdc-select__native-control')[0]
        elif name is 'label':
            return self.element.select('.mdc-floating-label')[0]

    # ----------------------------------------------------------------------

    @classmethod
    def add_option(cls, element, label, value, selected=False, disabled=False):
        """"""
        option = html.OPTION(label, value=value, selected=selected, disabled=disabled)
        cls['options_placeholder'] <= option
        return option

    # ----------------------------------------------------------------------

    @classmethod
    def add_options(cls, element, options, selected=None):
        """"""
        if selected is None:
            cls.add_option(element, '', '', selected=True)
        else:
            cls['label'].class_name += ' mdc-floating-label--float-above'

        for option in options:
            cls.add_option(element, *option, option[1] == selected)


########################################################################
class MDCSlider(MDCTemplate):
    """"""
    NAME = 'slider', 'MDCSlider'

    MDC_optionals = {

        'disabled': 'aria-disabled="true"',
        # 'box': 'mdc-select--box',

    }

    # ----------------------------------------------------------------------

    def __new__(self, label, min=0, max=100, step=1, valuenow=0, discrete=False, continuous=False, disabled=False, markers=False, **kwargs):
        """"""

        if not discrete and not continuous:
            continuous = True

        self.element = self.render(locals(), kwargs)
        # self.element.need_label = False

        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""

        cls.ID = cls.new_id()
        context['id'] = cls.ID

        if context['disabled']:
            context['input_disabled'] = 'disabled'
        else:
            context['input_disabled'] = ''

        if context['markers']:
            context['display_marker'] = 'mdc-slider--display-markers'
            context['marker_container'] = '<div class="mdc-slider__track-marker-container"></div>'
        else:
            context['display_marker'] = ''
            context['marker_container'] = ''

        if context.get('continuous'):
            code = """
                <div class="mdc-slider" tabindex="0" role="slider"
                     aria-valuemin="{min}" aria-valuemax="{max}" data-step="{step}" aria-valuenow={valuenow}"
                     aria-label="Select Value"
                     {disabled}>
                  <div class="mdc-slider__track-container">
                    <div class="mdc-slider__track"></div>
                  </div>
                  <div class="mdc-slider__thumb-container">
                    <svg class="mdc-slider__thumb" width="21" height="21">
                      <circle cx="10.5" cy="10.5" r="7.875"></circle>
                    </svg>
                    <div class="mdc-slider__focus-ring"></div>
                  </div>
                </div>
            """
        elif context.get('discrete'):
            code = """
                <div class="mdc-slider mdc-slider--discrete {display_marker}" tabindex="0" role="slider"
                     aria-valuemin="{min}" aria-valuemax="{max}" data-step="{step}" aria-valuenow="{valuenow}"
                     aria-label="Select Value">
                  <div class="mdc-slider__track-container">
                    <div class="mdc-slider__track"></div>
                    {marker_container}
                  </div>
                  <div class="mdc-slider__thumb-container">
                    <div class="mdc-slider__pin">
                      <span class="mdc-slider__pin-value-marker"></span>
                    </div>
                    <svg class="mdc-slider__thumb" width="21" height="21">
                      <circle cx="10.5" cy="10.5" r="7.875"></circle>
                    </svg>
                    <div class="mdc-slider__focus-ring"></div>
                  </div>
                </div>
            """

        return cls.render_html(code, context)


########################################################################
class MDCSwitch(MDCTemplate):
    """"""
    NAME = 'switchControl', 'MDCSwitch'

    MDC_optionals = {

        'disabled': 'disabled',
        # 'box': 'mdc-select--box',

    }

    # ----------------------------------------------------------------------

    def __new__(self, label, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""

        cls.ID = cls.new_id()
        context['id'] = cls.ID

        code = """
        <div class="mdc-switch">
          <div class="mdc-switch__track"></div>
          <div class="mdc-switch__thumb-underlay">
            <div class="mdc-switch__thumb">
                <input type="checkbox" id="{id}" class="mdc-switch__native-control" role="switch">
            </div>
          </div>
        </div>
        """

        return cls.render_html(code, context)


########################################################################
class MDCTextField(MDCTemplate):
    """"""
    NAME = 'textField', 'MDCTextField'

    MDC_optionals = {

        'disabled': 'mdc-text-field--disabled',
        # 'box': 'mdc-select--box',
        # 'fullwidth': 'mdc-text-field--fullwidth',

        'value': 'value="{value}"',


    }

    # ----------------------------------------------------------------------

    def __new__(self, label, value=False, type='text', leading_icon=False, trailing_icon=False, helper_text=False, helper_text_persistent=True, outlined=False, disabled=False, fullwidth=False, **kwargs):
        """"""

        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""
        cls.ID = cls.new_id()
        context['id'] = cls.ID

        if context['disabled']:
            context['input_disabled'] = 'disabled'
        else:
            context['input_disabled'] = ''

        if context['value']:
            context['upgraded'] = 'mdc-text-field--upgraded'
            context['float_above'] = 'mdc-floating-label--float-above'
        else:
            context['upgraded'] = ''
            context['float_above'] = ''

        if context['helper_text']:
            context['helper_text'] = 'aria-controls="username-helper-text" aria-describedby="username-helper-text"'
        else:
            context['helper_text'] = ''

        if context['leading_icon']:
            context['icon'] = '<i class="material-icons mdc-text-field__icon" tabindex="0" role="button">{icon}</i>'.format(icon=context['leading_icon'])
            context['leading_icon'] = 'mdc-text-field--with-leading-icon'

        elif context['trailing_icon']:
            context['icon'] = '<i class="material-icons mdc-text-field__icon" tabindex="0" role="button">{icon}</i>'.format(icon=context['trailing_icon'])
            context['trailing_icon'] = 'mdc-text-field--with-trailing-icon'

        else:
            context['leading_icon'] = ''
            context['trailing_icon'] = ''
            context['icon'] = ''

        if context.get('outlined'):
            code = """
                <div class="mdc-text-field mdc-text-field--outlined {disabled} {upgraded} {leading_icon} {trailing_icon}">
                    {icon}
                  <input type="{type}" id="{id}" class="mdc-text-field__input" {value}>
                  <div class="mdc-notched-outline">
                    <div class="mdc-notched-outline__leading"></div>
                    <div class="mdc-notched-outline__notch">
                      <label for="{id}" class="mdc-floating-label {float_above}">{label}</label>
                    </div>
                    <div class="mdc-notched-outline__trailing"></div>
                  </div>
                </div>
            """

        elif context.get('fullwidth'):
            code = """
            <div class="mdc-text-field mdc-text-field--fullwidth {upgraded} {disabled}">
              <input class="mdc-text-field__input" id="{id}" type="{type}" {value} placeholder="{label}" aria-label="{label}" {helper_text}>              
            </div>
            """

        else:
            code = """
                <div class="mdc-text-field {disabled} {upgraded} {leading_icon} {trailing_icon}">
                    {icon}
                  <input type="{type}" id="{id}" class="mdc-text-field__input" {input_disabled} {value} {helper_text}>
                  <label class="mdc-floating-label" for="{id}">{label}</label>
                  <div class="mdc-line-ripple"></div>
                </div>
            """

        return cls.render_html(code, context)


########################################################################
class MDCTextAreaField(MDCTemplate):
    """"""
    NAME = 'textField', 'MDCTextField'

    MDC_optionals = {

        'disabled': 'mdc-text-field--disabled',
        # 'dense': 'mdc-text-field--dense',
        # 'box': 'mdc-select--box',
        'fullwidth': 'mdc-text-field--fullwidth',

        # 'value': 'value="{value}"',


    }

    # ----------------------------------------------------------------------

    def __new__(self, label, value='', rows='2', cols='40', helper_text=False, helper_text_persistent=True, disabled=False, fullwidth=False, **kwargs):
        """"""

        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""
        cls.ID = cls.new_id()
        context['id'] = cls.ID

        if context['disabled']:
            context['input_disabled'] = 'disabled'
        else:
            context['input_disabled'] = ''

        if context['helper_text']:
            context['helper_text'] = 'aria-controls="username-helper-text" aria-describedby="username-helper-text"'
        else:
            context['helper_text'] = ''

        code = """
        <div class="mdc-text-field mdc-text-field--textarea {fullwidth} {disabled}">
          <textarea id="{id}" class="mdc-text-field__input" rows="{rows}" cols="{cols}" {helper_text} {input_disabled}>{value}</textarea>
          
            <div class="mdc-notched-outline">
              <div class="mdc-notched-outline__leading"></div>
              <div class="mdc-notched-outline__notch">
                <label for="{id}" class="mdc-floating-label">{label}</label>
              </div>
              <div class="mdc-notched-outline__trailing"></div>
            </div>

        </div>
        """

        return cls.render_html(code, context)



"""
Brython MDCFramework: Base
==========================



"""


from browser import document, window, html, timer, load

from .mdc.MDCDrawer import MDCDrawer

#from .core import MDC

# from functools import wraps
# from importlib import import_module


########################################################################
class MDCBase:
    """
    BASE

    """

    # ----------------------------------------------------------------------
    def __init__(self, preload=False, preload_timeout=5000, save_static=False, *args, **kwargs):
        """"""
        self.__last_view__ = None

        self.save_static = save_static

        self.close_drawer_on_view = True

        self.registered_views = {}
        # self.all_views = []
        self.register_class = kwargs
        # self.

        self.build()

        # t = Thread(target=self.timeout_preload)
        if preload:
            timer.set_timeout(self.__timeout_preload__, preload_timeout)
        # t.start()

    # ----------------------------------------------------------------------
    def generate_drawer(self, **kwargs):
        """"""
        self.drawer = MDCDrawer(**kwargs)

        if kwargs.get('mode', 'modal') == 'modal':
            self.__drawer_placeholder__.after(html.DIV(Class="mdc-drawer-scrim"))
            self.__drawer_placeholder__.after(self.drawer)
            self.drawer.attach()
        else:
            self.__drawer_placeholder__.after(self.drawer)

        for key in self.register_class:
            _, name, icon = self.register_class[key]
            item = self.drawer.mdc.add_item(name, icon, id=key, link='{}.html'.format(key), ignore_link=True)
            item.bind('click', self.__set_view__(key, item))

        self.drawer.mdc['items'][0].class_name += ' mdc-list-item--activated'
        [item.bind('click', self.__set_focus__) for item in document.select('.mdc-list-item')]

    # ----------------------------------------------------------------------

    def __set_view__(self, key, item):
        """"""
        def inset(evt):
            self.view(key)
        return inset

    # ----------------------------------------------------------------------

    def __set_focus__(self, element):
        """"""
        for item in document.select('.mdc-list-item'):
            item.class_name = item.class_name.replace('mdc-list-item--activated', '')
        try:
            element.target.class_name += ' mdc-list-item--activated'
        except:
            element.class_name += ' mdc-list-item--activated'

    # ----------------------------------------------------------------------

    @classmethod
    def load_styles(self, styles_list):
        """"""
        document.select('head')[0] <= [html.LINK(href='/static/{}'.format(style), type='text/css', rel='stylesheet') for style in styles_list]
        # [load('/static/{}'.format(style)) for style in styles_list]

    # ----------------------------------------------------------------------

    @classmethod
    def load_scripts(self, scripts_list):
        """"""
        # document.select('head')[0] <= [html.SCRIPT(src='/static/{}'.format(script), type='text/javascript') for script in scripts_list]
        # [load('/static/{}'.format(script)) for script in scripts_list]
        for script in scripts_list:
            if script.startswith('http'):
                load(script)
            else:
                load('/static/{}'.format(script))

    # ----------------------------------------------------------------------

    def build(self):
        """"""
        body = html.DIV(Class='radiant-main', style={'display': 'inline-flex', 'width': '100%', 'height': '100%', })

        self.__drawer_placeholder__ = html.SPAN()
        body <= self.__drawer_placeholder__

        self.container = html.DIV(Class="mdc-drawer-app-content", style={'width': '100vw'})
        body <= self.container

        document <= body

    # #----------------------------------------------------------------------
    # def register(self, class_, icon, name):
        # """"""
        # self.register_class[name] = (class_, icon)

    # ----------------------------------------------------------------------
    def remove_parent(self, element):
        """"""
        new_parent = element.parent
        new = element.child[0]

    # ----------------------------------------------------------------------
    def __timeout_preload__(self):
        """"""
        # print(self.register_class)
        for view in self.register_class:
            self.preload(view)

    # ----------------------------------------------------------------------
    # @coroutine

    def preload(self, name):
        """"""
        if name in self.registered_views:
            # print('already loaded')
            return

        else:
            #view = eval(name)
            view, _, _ = self.register_class[name]
            mod, class_ = view.split('.')
            mod = __import__(mod)
            view = getattr(mod, class_)
            view = view(self, preload=True)
            # view.onload()
            # self.container.clear()
            # self.container <= view.container
            self.registered_views[name] = view

    # ----------------------------------------------------------------------

    def view(self, name, fn=None, kwargs={}):
        """"""
        if self.__last_view__:
            self.registered_views[self.__last_view__].onchange()
            self.registered_views[self.__last_view__].onchange_()

        # print(name)
        if name in self.registered_views:
            self.container.clear()
            self.container <= self.registered_views[name].container
            if fn:
                getattr(self.registered_views[name], fn)(**kwargs)
            self.registered_views[name].onload()
            self.registered_views[name].onload_()
            self.registered_views[name].apply_styles()
            # self.registered_views[name].loaded()
            self.registered_views[name].firstload()

        else:
            #view = eval(name)
            view, _, _ = self.register_class[name]
            mod, class_ = view.split('.')
            mod = __import__(mod)
            view = getattr(mod, class_)
            view = view(self)
            view.name__ = name
            view.onload()
            view.onload_()
            view.apply_styles()
            self.container.clear()
            self.container <= view.container
            self.registered_views[name] = view
            # view.loaded()

        self.__last_view__ = name
        self.secure_load()
        self.secure_styles()

        self.__set_focus__(document.select('.mdc-list-item#{}'.format(name))[-1])

        if self.save_static:
            from radiant import Exporter
            html_code = document.select('html')[-1].html
            ex = Exporter()
            ex.export('{}.html'.format(name), html_code)

    # ----------------------------------------------------------------------

    def secure_load(self):
        """"""
        try:
            topappbar = window.mdc.topAppBar.MDCTopAppBar.attachTo(document.querySelector('.mdc-top-app-bar'))
            if hasattr(self, 'drawer') and document.querySelector('.mdc-top-app-bar').autodrawer:
                # document.select('.mdc-top-app-bar')[0].unbind('MDCTopAppBar:nav')
                # document.select('.mdc-top-app-bar')[0].bind('MDCTopAppBar:nav', lambda ev:self.drawer.mdc.toggle())
                document.select('.mdc-top-app-bar .mdc-top-app-bar__navigation-icon')[0].unbind()
                document.select('.mdc-top-app-bar .mdc-top-app-bar__navigation-icon')[0].bind('click', lambda ev: self.drawer.mdc.toggle())

        except:
            pass

        # [window.mdc.ripple.MDCRipple.attachTo(ripple) for ripple in document.select('[data-mdc-auto-init=MDCRipple]')]
        # [window.mdc.ripple.MDCRipple.attachTo(surface) for surface in document.select('.mdc-button')]
        # [window.mdc.ripple.MDCRipple.attachTo(surface) for surface in document.select('.mdc-ripple-surface')]

        try:
            [window.mdc.slider.MDCSlider.attachTo(slider) for slider in document.select('.mdc-slider')]
        except:
            pass

        # try:
            # [window.mdc.floatingLabel.MDCFloatingLabel.attachTo(label) for label in document.select('.mdc-floating-label')]
        # except:
            # pass
        # try:
            # [window.mdc.helperText.MDCTextFieldHelperText.attachTo(label) for label in document.select('.mdc-text-field-helper-text')]
        # except:
            # pass

        # try:
            # [window.mdc.lineRipple.MDCLineRipple.attachTo(label) for label in document.select('.mdc-line-ripple')]
        # except:
            # pass

        # try:
            # [window.mdc.textField.MDCTextField.attachTo(label) for label in document.select('.mdc-text-field')]
        # except:
            # pass

        if hasattr(self, 'drawer') and self.close_drawer_on_view:
            self.drawer.mdc.close()

    # ----------------------------------------------------------------------

    def secure_styles(self):
        """"""

        # document <= html.STYLE('.mdc-text-field__input {height: unset;}')


    #----------------------------------------------------------------------
    def hide(self):
        """"""
        document.select_one('.radiant-main').style = {'display': None,}


    #----------------------------------------------------------------------
    def show(self, mode="inline-flex"):
        """"""
        document.select_one('.radiant-main').style = {'display': mode,}


########################################################################
class MDCView:
    """"""

    # ----------------------------------------------------------------------
    def __init__(self, parent, preload=False):
        """Constructor"""

        self.main = parent

        if not preload:
            # self.main.mdc_drawer.mdc.close()
            self.main.container.clear()

        self.container = self.build()
            # self.connect()

    # ----------------------------------------------------------------------
    def firstload(self):
        """"""

    # ----------------------------------------------------------------------
    def onreloaded(self):
        """"""

    # ----------------------------------------------------------------------
    def onload(self):
        """"""

    # ----------------------------------------------------------------------
    def onchange(self):
        """"""

    # ----------------------------------------------------------------------
    def onchange_(self):
        """"""
        for e in document.select(f'.emtp-show-on-{self.name__}'):
            e.style = {'display': None, }

    # ----------------------------------------------------------------------
    def onload_(self):
        """"""
        for e in document.select(f'.emtp-show-on-{self.name__}'):
            e.style = {'display': 'flex', }

    # ----------------------------------------------------------------------
    def apply_styles(self):
        """"""
        styles = self.styles()
        for k in styles:
            for component in self.container.select(k):
                component.style = styles[k]

    # ----------------------------------------------------------------------
    def reloadview(self, event=None):
        """"""
        self.main.container.clear()
        self.container = self.build()
        self.onload()
        self.main.container <= self.container
        self.main.secure_load()
        self.onreloaded()

        # if fn:
            # fn(kwargs)

    # ----------------------------------------------------------------------

    def toggleclass(self, chip, class_):
        """"""
        if class_ in chip.class_name:
            chip.class_name = chip.class_name.replace(class_, '')
        else:
            chip.class_name += ' {}'.format(class_)

    # ----------------------------------------------------------------------
    @classmethod
    def subview(cls, view):
        """"""
        from functools import wraps

        @wraps(view)
        def wrapped(self, *args, **kwargs):
            """"""
            container = view(self)

            self.container.clear()
            self.container <= container

            self.main.secure_load()

            # self.main.container.clear()
            # self.container.clear()
        return wrapped

    # ----------------------------------------------------------------------
    def styles(self):
        """"""
        return {}


    # ----------------------------------------------------------------------
    def set_title(self, title):
        """"""
        document.select_one('body title').text = title


    # #----------------------------------------------------------------------
    # def open_link(self, link):
        # """"""
        # def inset(event):
            # event.preventDefault()
            # androidmain.open_url(link.href)
        # return inset


#########################################################################
# class htmlElement:
    # """"""

    # ----------------------------------------------------------------------
    # def __new__(self, element):
        # """"""
        #element.__getattr__ = self.__getattr__
        # return element

    # ----------------------------------------------------------------------
    # def __getattr__(self, attr):
        # """"""
        #name = self.getAttribute('mdc-name')

        # if attr is 'mdc':
            # return MDC.__mdc__(name, element=self)

        # elif attr is 'Foundation':
            # return MDC.__mdc__(name, element=self).mdc.foundation_





"""
=========
WSHandler
=========

WebSockets are not standard HTTP connections. The "handshake" is HTTP,
but after the handshake, the protocol is message-based.

This implementation suppose that input message are in JSON format, each input
message will trigger a response message (in JSON format too) with a ``type``
information.
"""

import json
import traceback
import logging

# from tornado.ioloop import IOLoop
from tornado import websocket
# from datetime import timedelta

from openbci import acquisition

from tornado import gen
# import numpy as np


########################################################################
class __WebSocketHandler(websocket.WebSocketHandler):
    """"""

    # ----------------------------------------------------------------------
    def initialize(self, bci_config):
        """Load config values from main app.

        Parameters
        ----------
        bci_config: dict
            Dictionary from runner.
        """

    # ----------------------------------------------------------------------
    def check_origin(self, origin):
        """Prevents a WebSocket from being used by another website that the user is also visiting."""
        return True

    # ----------------------------------------------------------------------
    def open(self, *args, **kwargs):
        """Return a message to client for confirm connection."""

        logging.warning(f'Adding client {self}')
        self.STATUS["CLIENTS"].append(self)

    # ----------------------------------------------------------------------
    def on_close(self):
        """Remove client from stream."""

        if self in self.STATUS["CLIENTS"]:
            self.STATUS["CLIENTS"].pop(self.STATUS["CLIENTS"].index(self))
        logging.info(f"Removing client {self}")

    # ----------------------------------------------------------------------
    @property
    def bci_device(self):
        """"""
        if hasattr(self, 'device_id'):
            return self.STATUS['DEVICES'][self.device_id]
        else:
            return None

    # ----------------------------------------------------------------------
    def on_message(self, message):
        """Process input message.

        Parameters
        ----------
        message: dict
            Dictionary with 'command' key as python method.
        """

        if not self in self.STATUS["CLIENTS"]:
            return

        request = json.loads(message)

        if not request.get('command', False):
            error = f'No argument "command" in this request./nmessage: {message}'
            logging.warning(error)
            self.send_data({'type': 'ERROR', 'message': error}, request)
            return

        command = request.get('command')
        kwargs = request.get('kwargs')

        if getattr(self.bci_device, command, False):  # command is a method
            try:
                logging.info("Running command: \"{}\"".format(command))
                response = getattr(self.bci_device, command)(**kwargs)  # call method
                self.send_data({'type': 'COMMAND', 'command': command, 'response': str(response), }, request)
            except AttributeError as e:
                t = {}
                t['message'] = str(e)
                t['traceback'] = traceback.format_exc()
                self.send_data({'type': 'ERROR', 'message': t, }, request)

        elif getattr(self, command, False):  # command is a method
            try:
                logging.info("Running command: \"{}\"".format(command))
                getattr(self, command)(**kwargs)  # call method
            except AttributeError as e:
                t = {}
                t['message'] = str(e)
                t['traceback'] = traceback.format_exc()
                self.send_data({'type': 'ERROR', 'message': t, }, request)
        else:
            self.send_data({'type': 'ERROR', 'message': 'No argument "command" in this request.', }, request)

    # ----------------------------------------------------------------------
    # @gen.coroutine
    def send_data(self, data, request={}):
        # global STREAMING, CLIENTS
        """Return data to client.

        Parameters
        ----------
        data: dict
            Dictionary with desired data to return.
        request: dict
            Dictionary with request parameters, like command.
        """

        try:  # send data to client
            for client in self.STATUS["CLIENTS"]:
                client.write_message(data)
            logging.info(f'Writing data {data.get("pack_id", "")} for {len(self.STATUS["CLIENTS"])} clients, {data["buffer_pack"]}')
        except Exception as e:
            pass

        except:
            pass


    # ----------------------------------------------------------------------
    @gen.coroutine
    def timed_stream(self):
        """Prepare package for send it via ws."""


        # if self.STATUS["CLIENTS"] and self.STATUS["STREAMING"]:
            # # pass
            # # If clients connected call again.
            # # IOLoop.instance().add_timeout(timedelta(seconds=0.1), self.timed_stream)
            # IOLoop.instance().call_later(0.1, self.timed_stream)
        # else:
            # # If no clients stop streamming
            # if not self.STATUS["CLIENTS"]:
                # logging.warning('No clients!!')
            # elif not self.STATUS["STREAMING"]:
                # logging.warning('No streamming!!')
            # self.STATUS["STREAMING"] = False
            # self.stop_stream()
            # return

        # if pack_size := self.bci_device.eeg_pack.qsize():
            # # return

            # data = {
                # 'type': 'STREAM',
                # 'pack_id': self.STATUS["PACK_ID"],
                # 'buffer_eeg': self.bci_device.eeg_buffer.qsize(),
                # 'buffer_pack': pack_size - 1,
                # 'buffer_serial': len(self.bci_device.eeg_serial),
            # }

            # if not self.STATUS["CLEANED"] and self.STATUS["PACK_ID"] == 0:
                # self.bci_device.reset_buffers()
                # # self.STATUS["PACK_ID"] = 0
                # self.STATUS["CLEANED"] = False

            # data['data'] = [d.tolist() for d in self.bci_device.eeg_pack.get()]
            # self.send_data(data)
            # self.STATUS["PACK_ID"] += 1




        while self.STATUS["CLIENTS"] and self.STATUS["STREAMING"]:
            # pass
            # If clients connected call again.
            # IOLoop.instance().add_timeout(timedelta(seconds=0.1), self.timed_stream)
            # IOLoop.instance().call_later(0.1, self.timed_stream)

            if pack_size := self.bci_device.eeg_pack.qsize():

                if pack_size > 2:
                    self.bci_device.eeg_pack.get()

                data = {
                    'type': 'STREAM',
                    'pack_id': self.STATUS["PACK_ID"],
                    'buffer_eeg': self.bci_device.eeg_buffer.qsize(),
                    'buffer_pack': self.bci_device.eeg_pack.qsize() - 1,
                    # 'buffer_serial': len(self.bci_device.eeg_serial),
                }

                if not self.STATUS["CLEANED"] and self.STATUS["PACK_ID"] == 0:
                    self.bci_device.reset_buffers()
                    # self.STATUS["PACK_ID"] = 0
                    self.STATUS["CLEANED"] = False
                    yield


                # dta = []
                # for _ in range(pack_size):
                    # dta.append([d.tolist() for d in self.bci_device.eeg_pack.get()])
                # dta = np.array(dta)
                # data['data'] = [np.concatenate(dta[:, i]).tolist() for i in range(dta.shape[1])]

                data['data'] = [d.tolist() for d in self.bci_device.eeg_pack.get()]

                self.send_data(data)
                self.STATUS["PACK_ID"] += 1

                yield

            yield


        # If no clients stop streamming
        if not self.STATUS["CLIENTS"]:
            logging.warning('No clients!!')
        elif not self.STATUS["STREAMING"]:
            logging.warning('No streamming!!')

        self.stop_stream()
        return




########################################################################
class __WSHandler(__WebSocketHandler):
    """Websocket handler."""

    # Attributes in a mutable object.
    STATUS = dict(

        CLIENTS=[],
        STREAMING=False,
        PACK_ID=-5,
        CLEANED=False,
        DEVICES={},

    )

    # ----------------------------------------------------------------------
    def register(self, device_id, **kwargs):
        """"""
        if device_id in self.STATUS['DEVICES']:
            # self.device_id = device_id
            logging.info(f"Client {self} registered for {device_id} device")

        else:
            logging.warning(f"No device {device_id} attached.")
            self.set_up(device_id, **kwargs)


    # ----------------------------------------------------------------------
    def start_stream(self, milliseconds, boardmode=None):
        """Start streamming in device with selected boardmode,"""

        if (not self.STATUS["STREAMING"]):# or (not self.bci_device is None):
            self.STATUS["STREAMING"] = True
            if boardmode:
                self.bci_device.command(getattr(self.bci_device, boardmode))
            self.bci_device.start_collect(milliseconds=milliseconds)

            logging.warning('Start streamming')
            self.timed_stream()
            # self.timed_stream_callback = PeriodicCallback(self.timed_stream, callback_time=milliseconds, jitter=0.1)
            # self.timed_stream_callback.start()
        else:
            logging.warning('Already streamming')
            self.stop_stream()
            self.start_stream(milliseconds, boardmode)

    # ----------------------------------------------------------------------
    def stop_stream(self, **request):
        """Stop streamming in device."""

        self.STATUS["STREAMING"] = False
        logging.warning('Stop streamming')
        self.bci_device.stop_collect()
        # self.timed_stream_callback.stop()


########################################################################
class WSHandler_Serial(__WSHandler):
    """`bci_device` is created explicitly.

    This means that RFduino must be placed before start the system.
    """

    # ----------------------------------------------------------------------
    def set_up(self, device_id=None, montage=None, sample_rate='SAMPLE_RATE_250HZ'):
        """"""
        if device_id in self.STATUS['DEVICES']:
            bci_device = self.STATUS['DEVICES'][device_id]
            bci_device.close()

        try:
            bci_device = acquisition.CytonRFDuino(device_id, montage, sample_rate)
            self.STATUS['DEVICES'][device_id] = bci_device
            self.device_id = device_id
            logging.info(f"Client {self} registered for {device_id} device")

        except:
            logging.warning(f"No device attached: {e}")


########################################################################
class WSHandler_WiFi(__WSHandler):
    """`bci_device` is created explicitly.

    This means that WiFi shield must be connected to the AP or network.

    """
    # ----------------------------------------------------------------------
    def set_up(self, device_id, montage=None, local_ip_address=None, sample_rate='SAMPLE_RATE_250HZ'):
        """"""
        if device_id in self.STATUS['DEVICES']:
            bci_device = self.STATUS['DEVICES'][device_id]
            bci_device.close()

        try:
            bci_device = acquisition.CytonWiFi(device_id, montage, sample_rate)
            self.STATUS['DEVICES'][device_id] = bci_device
            self.device_id = device_id
            logging.info(f"Client {self} registered for {device_id} device")

        except Exception as e:
            logging.warning(f"No device attached: {e}")




from browser import window

from .futures import Future


MODE_BINARY = 0
MODE_TEXT = 1
MODE_DATAURL = 2

class FileFuture(Future):
    
    def __init__(self, file, mode=MODE_TEXT):
        super().__init__()
        self._reader = window.FileReader.new()
        self._reader.bind('load', self._load_handler)
        self._mode = mode
        if self._mode == MODE_BINARY:
            raise NotImplementedError()
        if mode == MODE_BINARY:
            self._reader.readAsBinaryString(file)
        elif mode == MODE_TEXT:
            self._reader.readAsText(file)
        elif mode == MODE_DATAURL:
            self._reader.readAsDataURL(file)
        
    def _load_handler(self, evt):
        self.set_result(evt.target.result)
        
        


def read(file, mode):
    return FileFuture(file, mode)



import os
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
   README = readme.read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))
version_str = open(os.path.join('openbci', '_version.txt'), 'r').read().strip()

setup(
    name='openbci',
    version=version_str,
    packages=['openbci', 'openbci.acquisition', 'openbci.preprocess'],

    author='Yeison Cardona',
    author_email='yeisoneng@gmail.com',
    maintainer='Yeison Cardona',
    maintainer_email='yeisoneng@gmail.com',

    # url='http://yeisoncardona.com/',
    download_url='https://bitbucket.org/gcpds/openbci/downloads/',

    install_requires=['pyserial',
                      'scipy>=1.3.1',
                      'numpy',
                      'psutil',
                      'mne',
                      'requests',
                      'tornado',
                      ],


    extras_require={
        'utils': ['systemd_service', 'netifaces', 'python-nmap', 'quickapi'],
    },

    scripts=[
        "cmd/eegstream",
    ],

    include_package_data=True,
    license='Simplified BSD License',
    description="Hight level Python module for OpenBCI hardware by GCPDS.",
    # long_description="",

    long_description=README,
    long_description_content_type='text/markdown',


    python_requires='>=3.6',

    classifiers=[
       'Development Status :: 4 - Beta',
       'Intended Audience :: Developers',
       'Intended Audience :: Education',
       'Intended Audience :: Healthcare Industry',
       'Intended Audience :: Science/Research',
       'License :: OSI Approved :: BSD License',
       'Programming Language :: Python :: 3.7',
       'Programming Language :: Python :: 3.8',
       'Topic :: Scientific/Engineering',
       'Topic :: Scientific/Engineering :: Artificial Intelligence',
       'Topic :: Scientific/Engineering :: Human Machine Interfaces',
       'Topic :: Scientific/Engineering :: Medical Science Apps.',
       'Topic :: Software Development :: Embedded Systems',
       'Topic :: Software Development :: Libraries',
       'Topic :: Software Development :: Libraries :: Application Frameworks',
       'Topic :: Software Development :: Libraries :: Python Modules',
       'Topic :: System :: Hardware :: Hardware Drivers',
    ],

)



"""
    The webworker module provides a basic integration between Brython and
    [WebWorkers](https://developer.mozilla.org/en-US/docs/Web/API/Web_Workers_API)).
    It allows one to run a python script in a web worker with relative ease. Currently
    there are two classes which can be used. A basic Worker class and a RPCWorker class.
"""
from browser import window


print("~"*100)

import asyncio
import os
import sys

DEFAULT_BRYTHON_OPTIONS = getattr(__BRYTHON__, '$options')

CHILD_WORKERS = []

# The following are possible states
# the worker can be in throughout
# its life

S_CREATED = 0       # The worker has been created (e.g. new Worker())
S_LOADING = 1       # The webworker started loading brython
S_LOADED = 2        # The webworker finished loading brython and
                    # is now starting to run the python script
S_RUNNING = 3       # The python script finished its initialization
                    # and is ready to work
S_FINISHED = 4      # The worker finished without error (exit(0))
S_TERMINATED = 5    # The worker was terminated

try:
    _Worker = window.Worker.new
    _can_launch_workers = True
except:
    _can_launch_workers = False

class WorkerError(Exception):
    pass


class Message:
    """
        A class representing a message to be sent to/from the webworker.
        It has a name (message type) and some data (message payload).
        These are mandatory arguments to be passed to the constructor.
    """
    def __init__(self, name, data, src=None, id=-1):
        self.name = name
        self.data = data
        self.id = id
        self.is_reply = False
        self.src = None

    def __str__(self):
        return "MSG("+str(self.name)+":"+str(self.id)+"):"+str(self.data)

class Reply(asyncio.Future):
    """
        A future representing a reply to a message. Not to be used directly
        by the user but is created, e.g., by the `post_message` method when
        called with `want_reply=True`.
    """
    _LAST_MESSAGE_ID = 0
    _WAITING_REPLIES = {}

    @classmethod
    def _next_id(cls):
        cls._LAST_MESSAGE_ID +=1
        return cls._LAST_MESSAGE_ID

    @classmethod
    def terminate(cls, worker, reason=None):
        wr = {}
        text = "Worker Terminated"
        if reason is not None:
            text += "\n" + str(reason)
        for id, reply in cls._WAITING_REPLIES.items():
            if reply._worker == worker:
                reply.set_exception(WorkerError(text))
            else:
                wr[id] = reply
        cls._WAITING_REPLIES = wr

    def __init__(self, message, timeout, worker=None):
        super().__init__()
        message.id = self._next_id()
        self._wait_id = message.id
        self._WAITING_REPLIES[message.id] = self
        self._worker = worker
        self.add_done_callback(self.finish_waiting)
        if timeout:
            self._timeout = self._loop.call_later(timeout, self.set_exception, asyncio.TimeoutError())
        else:
            self._timeout = None

    def finish_waiting(self, *args, **kwargs):
        if self._timeout:
            self._timeout.cancel()

        if self._wait_id in self._WAITING_REPLIES:
            del self._WAITING_REPLIES[self._wait_id]

    def set_result(self, result):
        super().set_result(result)

class WorkerCommon:
    """
        This class implements methods useful both in the main thread
        and in the worker thread.
    """


    @property
    def status(self):
        """
            The status of the worker:

               S_CREATED        - The worker has been created (e.g. new Worker())
               S_LOADING        - The webworker started loading brython
               S_LOADED         - The webworker finished loading brython and
                                  is now starting to run the python script
               S_RUNNING        - The python script finished its initialization
                                  and is ready to work
               S_FINISHED       - The worker finished without error (exit(0))
               S_TERMINATED     - The worker was terminated

        """
        return self._status

    def __init__(self, worker):
        self._status = S_CREATED
        self._event_handlers = {}
        self._message_handlers = {}
        self._worker = worker
        self._worker.addEventListener('message', self._event_handler)
        self._worker.addEventListener('error', self.__error_handler)
        self._queued_messages = []
        self.bind_event('message', self._message_handler)
        self.bind_event('ready', self._post_queued_messages)

    def post_message(self, message, want_reply=False, timeout=None):
        """
            Sends the message `message` (an instance of the `Message` class)
            to the other side. If `want_reply` is set to true, the method
            returns a future representing the potential reply to this message.
            The optional `timeout` parameter specifies how long in seconds the future
            should wait for a reply before resolving itself with a timeout exception,
        """
        if self.status > S_RUNNING:
            raise WorkerError("Invalid state")
        if want_reply:
            ret = Reply(message, timeout, worker=self)
        else:
            ret = None
        self._queued_messages.append({'type':'message', 'name':message.name, 'id':message.id, 'data':message.data})
        if self.status == S_RUNNING:
            self._post_queued_messages()
        return ret

    def post_reply(self, message, reply):
        """
            Sends the message `reply` as a reply to the message `message`.
        """
        payload = {'type':'message', 'name':reply.name, 'id':reply.id, 'data':reply.data}
        if message.id:
            payload['reply_to']=message.id
        self._queued_messages.append(payload)
        if self.status == S_RUNNING:
            self._post_queued_messages()

    def _post_queued_messages(self, *_, **kwargs):
        for payload in self._queued_messages:
            self._worker.postMessage(payload)
        self._queued_messages.clear()

    def bind_event(self, event, handler):
        """
            Registers `handler` to be called whenever the `event` is emitted
            by the class. Events are emitted when different things happen.
            For example when the state of the worker changes, the following
            events re emitted:

               exited
               loaded
               ready
        """
        self._bind(self._event_handlers, event, handler)

    def unbind_event(self, event=None, handler=None):
        self._unbind(self._event_handlers, event, handler)

    def bind_message(self, message, handler):
        """
            Registers the method `handler` to be called whenever a message of type `message`
            arrives.
        """
        self._bind(self._message_handlers, message, handler)

    def unbind_message(self, message=None, handler=None):
        self._unbind(self._message_handlers, message, handler)


    def _bind(self, handler_list, event, handler):
        handlers = handler_list.get(event, [])
        handlers.append(handler)
        handler_list[event] = handlers

    def _unbind(self, handler_list, event=None, handler=None):
        if event is None and handler is None:
            handler_list = {}
        elif handler is None:
            handler_list[event] = []
        elif event is None:
            for ev, handlers in handler_list.items():
                if handler in handlers:
                    handlers.remove(handler)
        else:
            handlers = handler_list.get(event, [])
            if handler in handlers:
                handlers.remove(handler)

    def _emit_event(self, event, data=None):
        self._emit(self._event_handlers, event, data)

    def _emit_message(self, message, msg):
        self._emit(self._message_handlers, message, msg)

    def _emit(self, handler_list, event, data=None):
        try:
            handlers = handler_list.get(event, [])
            for h in handlers:
                h(event, data, src=self)
        except Exception as ex:
            # TODO: Add error handling
            print("Exception", str(ex), "while handling", event, "data=", str(data.data))


    def __error_handler(self, error, *_, **kwargs):
        self._emit_event('error', error)

    def _event_handler(self, event, *_, **kwargs):
        try:
            event_type = event.data['type']
            self._emit_event(event_type, event.data)
        except:
            pass

    def _message_handler(self, event, data, *_, **kwargs):
        msg = Message(data['name'],data['data'], src=self, id=data['id'])
        if 'reply_to' in data:
            msg.is_reply = True
            reply = Reply._WAITING_REPLIES.get(data['reply_to'], None)
            if reply:
                reply.set_result(msg)
        else:
            self._emit_message(msg.name, msg)


class WorkerParent(WorkerCommon):
    """
        The class representing the worker in the main thread.
    """
    WORKER_SCRIPT = sys.base_exec_prefix+'/web_workers/worker.js'
    CHILD_CLASS = 'browser.webworker.WorkerChild'

    def __init__(self, url, argv=[], environ={}, brython_options=DEFAULT_BRYTHON_OPTIONS):
        """
            Creates a new Web Worker and runs the python script downloaded from `url`
            in it. The parameter `url` can be either relative to the directory where
            Brython is installed or it must be an absolute url (i.e. start with '/' or 'http:').

            The parameter `argv` is a list of arguments which will be available as `sys.argv` in the worker.
            The parameter `environ` is a dictionary which will be available as `os.environ` in the worker.
            The parameter `brython_options` will be passed to the `brython` function when called from
            the webworker. If it contains an `imports` key, this should be a list of javascript files
            to be loaded instead of the standard brython files. If it contains the key `import_dist`,
            it will load brython from the `brython_webworker_dist.js` script.  By default it uses the
            same options passed to the parent `brython` function.
        """
        if not _can_launch_workers:
            raise WorkerError("Cannot spawn workers (webkit based browsers don't support running webworkers inside webworkers)")
        super().__init__(_Worker(self.WORKER_SCRIPT,{"name":url}))
        self.bind_event('status', self._status_handler)
        self.bind_event('error', self._error_handler)
        self._status_waiters = []
        self._status = S_LOADING
        self._error = None
        self._worker.postMessage({
                'program':{'url':url},
                'brython_options':brython_options,
                'argv':[url]+argv,
                'env':environ,
                'worker_class':self.CHILD_CLASS
         })
        CHILD_WORKERS.append(self)

    def terminate(self):
        """
            Forcibly Terminates the worker.

            Raises an error if the worker is not running
        """
        if self.status > S_RUNNING:
            raise WorkerError("Invalid state")
        self._worker.terminate()
        self._status = S_TERMINATED
        Reply.terminate(self, self._error)
        self._emit_event('exited')

    def wait_for_status(self, status):
        """
            Returns a future which will be resolved when
            the worker transitions to the given status (or any later status)
        """
        fut = asyncio.Future()
        setattr(fut, 'waiting_for', status)
        if status <= self.status:
            fut.set_result(True)
        else:
            self._status_waiters.append(fut)
        return fut


    def _status_handler(self, event, data, *_, **kwargs):
        self._status = data.status
        keep = []
        for f in self._status_waiters:
            if f.waiting_for <= self._status:
                f.set_result(True)
            else:
                keep.append(f)
        self._status_waiters = keep
        if self._status >= S_FINISHED:
            self._emit_event('exited')
            if 'error' in data:
                self._error = data.error
            Reply.terminate(self, self._error)
            CHILD_WORKERS.remove(self)
        elif self._status == S_LOADED:
            self._emit_event('loaded')
        elif self._status == S_RUNNING:
            self._emit_event('ready')

    def _error_handler(self, error, data, *_, **kwargs):
        self._error = data
        self.terminate()


class WorkerChild(WorkerCommon):
    """
        The class representing the worker in the worker thread. Should not
        be instantiated by the user directly. Instead an instance of this
        class will be provided as `current_worker` to every worker started
        by the `WorkerParent` class.
    """
    def __init__(self):
        super().__init__(__BRYTHON__._WORKER)
        self._argv = sys.argv
        self._environ = __BRYTHON__._ENV
        self._status = S_LOADED

    def exec(self):
        """
            This method should be called from the worker script when it is
            ready to start receiving messages from the main thread.
        """
        self._status = S_RUNNING
        self._worker.postMessage({'type':'status', 'status':S_RUNNING})
        self._emit_event('ready')

    def terminate(self):
        """
            The worker script can use this method to gracefully shutdown the worker.
        """
        if self.status > S_RUNNING:
            raise WorkerError("Invalid state")
        self._status = S_FINISHED
        self._emit_event('exited')
        self._worker.postMessage({'type':'status', 'status':S_FINISHED})
        self._worker.close()

class RPCWorkerParent(WorkerParent):
    CHILD_CLASS = 'browser.webworker.RPCWorkerChild'

    def __init__(self, url, argv=[], environ={}, brython_options=DEFAULT_BRYTHON_OPTIONS):
        super().__init__(url, argv, environ, brython_options)
        self.bind_message('register', self._register)

    def _register(self, msg_name, message, *_, **kwargs ):
        for method, doc in message.data:
            self._generate_method(method, doc)

    def _generate_method(self, method, doc):
        @asyncio.coroutine
        def meth(*args, **kwargs):
            msg = Message('call', {'method':method, 'args':args, 'kwargs':kwargs})
            reply = yield self.post_message(msg, want_reply=True)
            if reply.data['status'] == 'ok':
                return reply.data['ret']
            else:
                raise WorkerError(reply.data['error'])

        meth.__doc__ = doc
        meth.__name__ = method
        setattr(self, method, meth)

class RPCWorkerChild(WorkerChild):
    def __init__(self):
        super().__init__()
        self._methods = {}
        self.bind_message('call', self._call)

    def register_method(self, m):
        m_name = m.__name__.split('.')[-1]
        m_doc = m.__doc__ or ''
        self._methods[m_name] = m
        msg = Message('register', [(m_name, m_doc)])
        self.post_message(msg)

    def _call(self, msg_name, message, *_, **kw):

        m_name = message.data['method']
        args = message.data['args']
        kwargs = message.data['kwargs']
        if m_name not in self._methods:
            self.post_reply(message, Message('result', {'status':'error','error':'Method not registered'}))
        m = self._methods.get(m_name)
        try:
            ret = m(*args, **kwargs)
            msg = Message('result', {'status':'ok', 'ret':ret})
            self.post_reply(message, msg)
        except Exception as e:
            msg = Message('result', {'status':'error', 'error':"Exception raised in call:"+str(e)})
            self.post_reply(message, msg)

if __BRYTHON__.isa_web_worker:
    w_cls = __BRYTHON__._WORKER_CLASS
    elements = w_cls.split('.')
    cls_name = elements[-1]
    cls_module = '.'.join(elements[:-1])
    try:
        mod = __import__(cls_module, fromlist=[cls_name])
        current_worker = getattr(mod, cls_name)()
    except Exception as ex:
        print("Error importing child worker class", ex)
        __BRYTHON__._WORKER.postMessage({'type':'status', 'status':S_TERMINATED, 'error':str(ex)})
        __BRYTHON__._WORKER.close()
    #sys.argv = current_worker._argv
    os.environ.update(dict(current_worker._environ))
else:
    current_worker = None



from .coroutines import coroutine, ensure_future
from .futures import Future, gather
from ._utils import decorator

import logging
logger = logging.getLogger(__package__)

class AsyncObjectInitCanceledError(Exception):
    def __str__(self):
        return "Async object initialization canceled"

class AsyncObjectInitExceptionError(Exception):
    def __init__(self, ex):
        super().__init__("Init threw an exception:"+str(ex))
        self._init_ex = ex

class AsyncObjectUninitializedError(Exception):
    def __str__(self):
        return "Async object not initialized"

@decorator
def async_init(init):
    """
        A decorator for asynchronous constructors.
    """
    def new_init(self, *args, **kwargs):
        logger.debug("Calling decorated init for %s", self.__class__.__name__)
        cls = str(self.__class__)
        def _init_cb(fut):
            if fut.cancelled():
                logger.error("Constructor for %s canceled.", cls)
            elif fut.exception() is not None:
                print(fut.exception())
                logger.error("Constructor for %s raised an exception: %s", cls, str(fut.exception()))
        self._init_future = coroutine(init)(self, *args, **kwargs)
        self._init_future.add_done_callback(_init_cb)

    return new_init



def defer(future, func, *args, **kwargs):
    ret = Future()
    def cb(fut):
        if fut.canceled():
            logger.error("Unable to call deferred method %s object init canceled.", str(func.__decorated))
            ret.set_exception(AsyncObjectInitCanceledError())
        if fut.exception() is not None:
            logger.error("Unable to call deferred method %s object init raised an exception:", str(fut.exception))
            ret.set_exception(AsyncObjectInitExceptionError(fut.exception()))
        else:
            try:
                ret.set_result(func(*args, **kwargs))
            except Exception as ex:
                ret.set_exception(ex)
    future.add_done_callback(cb)
    return ret

@decorator
def _generate_guard(func):
    def guard(self, *args, **kwargs):
        if self._init_future.cancelled():
            logger.error("Unable to call deferred method %s object init canceled.", str(func))
            raise AsyncObjectInitCanceledError()
        elif self._init_future.exception() is not None:
            raise AsyncObjectInitExceptionError(self._init_future.exception())
        elif self._init_future.done():
            return func(self, *args, **kwargs)
        else:
            if hasattr(func, '__async_task__'):
                logger.info("Defering method %s until object is initialized.", str(func))
                return defer(self._init_future, func, self, *args, **kwargs)
            else:
                logger.error("Calling method on Uninitialized object")
                raise AsyncObjectUninitializedError()
    return guard


def async_class(cls):
    """
        A decorator for classes having an asynchronous constructor.
        Care must be given that all method invocations must be deferred until
        the constructor has finished. This is done by the :function:`_generate_guard`
        function which is called on each nonprivate class method.
    """
    for member in dir(cls):
        if not member[0:2] == '__':
            meth = getattr(cls, member)
            if hasattr(meth, '__call__'):
                setattr(cls, member, _generate_guard(meth))
    return cls


class FutureCall(Future):
    """
        A future representing the result of a method applied to a
        future result.
    """

    def __init__(self, fn, fut, loop=None):
        super().__init__(loop=loop)
        self._fn = fn
        ensure_future(fut, loop=loop).add_done_callback(self._done)

    def _done(self, fut):
        ex = fut.exception()
        if ex is None:
            res = self._fn(fut.result())
            if isinstance(res, Future):
                res.add_done_callback(self._done)
            else:
                self.set_result(res)
        else:
            self.set_exception(fut.exception())


def val(fut_or_val):
    """
        Utility function for handling finished Futures as values.
        If :param:`fut_or_val` is a Future, returns the future's
        result otherwise returns :param:`fut_or_val`.
    """
    if isinstance(fut_or_val, Future):
        return fut_or_val.result()
    return fut_or_val


def wrap_in_future(res, loop=None):
    """
        Unconditionally wraps a value in a future. It does this
        by calling asyncio.ensure_future and, if that fails,
        creating a dummy future and setting its result to
        :param:`res`.
    """
    try:
        return ensure_future(res, loop=loop)
    except TypeError:
        pass
    fut = Future()
    fut.set_result(res)
    return fut


def wait_for(*args, loop=None):
    """
        Returns a future waiting for all its arguments to recursively resolve
        to a non-future result.
    """
    return gather(*[FullyResolved(wrap_in_future(a, loop=loop)) for a in args])


class FullyResolved(Future):
    """
        Wraps a future into one which returns a recursively resolved result, i.e.
        if the future finishes and its result is another future, this class waits
        until the other future is also done, and so on.
    """
    def __init__(self, fut, loop=None):
        super().__init__(loop=loop)
        self._fut = wrap_in_future(fut, loop=loop)
        self._fut.add_done_callback(self._done)

    def _done(self, fut):
        if fut.exception() is not None:
            self.set_exception(fut.exception())
        else:
            try:
                self._fut = ensure_future(fut.result())
                self._fut.add_done_callback(self._done)
            except TypeError:
                self.set_result(fut.result())


class FutureObject(Future):
    """
        A future simulating a future object. I.e. one can do attribute access and method calls
        on the future objects with the result being a future which is resolved once the
        future object is resolved. Typical use is in functions which are not coroutines
        (or async) and so can't use yield from or await, but still want to deal with asynchronous
        calls in a reasonable way. E.g. assume we have a method to asynchronously open a file.
        In a normal async function we could write

        .. code-block:: python

            async def test():
                f = await async_open('/tmp/test.txt')
                f.write('test')
                f.close()

        The :class:`FutureObject` class allows writing similar code even in functions
        which are not async:

        .. code-block:: python

            def test():
                f = FutureObject(async_open('/tmp/test.txt'))
                f.write('test') # returns a future which we ignore it
                f.write(f.name) # also works (!)
                f.close()       # returns a future which we ignore

        Note: that the arguments of all "future" function calls are first resolved. This is
        for convenience's sake. However it means that methods which expect Futures as arguments
        wont work!
    """
    FORBIDDEN = [name for name in dir(Future) if not name.startswith('__')] + \
                ['_loop', '_callbacks'] +\
                ['__getattr__', '__setattr__', '__call__'] + \
                ['_fut', '_done']

    @classmethod
    def future_or_val(cls, x, loop=None):
        """
            If :param:`x` is a :class:`asyncio.Future` wraps it in a :class:`FutureObject`, otherwise
            returns :param:`x` unchanged.
        """
        try:
            return FutureObject(ensure_future(x, loop=loop))
        except TypeError:
            return x
        return x

    def __init__(self, fut, loop=None):
        super().__init__(loop=loop)
        self._fut = ensure_future(fut, loop=loop)
        self._fut.add_done_callback(self._done)

    def _done(self, fut):
        if fut.exception() is None:
            self.set_result(fut.result())
        else:
            self.set_exception(fut.exception())

    def __getattr__(self, name):
        if name in FutureObject.FORBIDDEN:
            return super().__getattribute__(name)

        if self.done():
            return self.future_or_val(getattr(self.result(), name), loop=self._loop)

        ret = Future()

        def resolve(fut):
            ex = fut.exception()
            if ex is None:
                ret.set_result(self.future_or_val(getattr(fut.result(), name), loop=self._loop))
            else:
                ret.set_exception(fut.exception())

        self._fut.add_done_callback(resolve)

        return FutureObject(ret)

    def __setattr__(self, name, value):
        if name in FutureObject.FORBIDDEN:
            return super().__setattr__(name, value)

        ready_fut = wait_for(self, value, loop=self._loop)

        if ready_fut.done():
            return self.future_or_val(setattr(self.result(), name, val(value)), loop=self._loop)

        ret = Future()

        def resolve(fut):
            ex = fut.exception()
            if ex is None:
                ret.set_result(self.future_or_val(setattr(self.result(), name, val(value)), loop=self._loop))
            else:
                ret.set_exception(fut.exception())

        self._fut.add_done_callback(resolve)

        return FutureObject(ret)

    def __call__(self, *args, **kwargs):

        ready_fut = wait_for(self, *args, *kwargs.values(), loop=self._loop)

        if ready_fut.done():
            return self.future_or_val(self.result()(*[val(a) for a in args], **{k: val(v) for k, v in kwargs.items()}), loop=self._loop)

        ret = Future()

        def resolve(fut):
            ex = fut.exception()
            if ex is None:
                ret.set_result(self.future_or_val(self.result()(*[val(a) for a in args], **{k: val(v) for k, v in kwargs.items()}), loop=self._loop))
            else:
                ret.set_exception(fut.exception())

        ready_fut.add_done_callback(resolve)

        return FutureObject(ret)






"""
================
Cyton Bin Parser
================
"""

import numpy as np
import logging

import os
import sys
from serial.serialutil import SerialException
import time
import struct
from abc import ABCMeta, abstractmethod
from datetime import datetime, timedelta

import threading
threading.stack_size(2**26)
sys.setrecursionlimit(2**20)

TIMEOUT_QUEUE = 0.01
SAMPLE_RATE_PACK = 50
# DROP_SAMPLES = 10
# DROP_PROPORTION = 1

from queue import Empty, deque, Queue
from threading import Thread


########################################################################
class CytonBinParser(metaclass=ABCMeta):
    """
    The main objective of this class (with inheritance check) create a
    background process for read and other for deserialize data and serve their
    content in queue buffers.
    """

    # Flags
    RAW = False
    READING = None
    BOARDMODE = None

    # ----------------------------------------------------------------------
    def __init__(self):
        """"""
        self._data_eeg = Queue()
        # self._data_serial = Queue()
        self._data_serial = deque()
        self._data_packs = Queue()

        # self._serial_buffer = Queue(maxsize=100)
        # [self._serial_buffer.put(0) for i in range(100)]

        self._unsampled = Queue(maxsize=3)
        [self._unsampled.put([0] * 8) for i in range(3)]

        self.stop()

        self._valid = 1
        self._no_valid = 1
        self._last_marker = 0
        self._drop_proportion = 1

        self.reset_input_buffer()

    # ----------------------------------------------------------------------
    @abstractmethod
    def start(self):
        """Device handled process, start data stream."""
        pass

    # ----------------------------------------------------------------------
    @abstractmethod
    def stop(self):
        """Device handled process, stop data stream."""
        pass

    # ----------------------------------------------------------------------
    @abstractmethod
    def read(self):
        """Device handled process, read binary data."""
        pass

    # ----------------------------------------------------------------------
    @abstractmethod
    def write(self):
        """"""
        """Device handled process, write bytes."""
        pass

    # ----------------------------------------------------------------------
    @abstractmethod
    def get_board_mode(self):
        """Device handled process, board mode."""
        pass

    # ----------------------------------------------------------------------
    @abstractmethod
    def scale_factor_eeg(self):
        """Device handled process, scale factor."""
        pass

    # ----------------------------------------------------------------------
    @abstractmethod
    def daisy(self):
        """Device handled process, daisy board."""
        pass

    # ----------------------------------------------------------------------
    @abstractmethod
    def reset_input_buffer(self):
        """Device handled process, flush input data."""
        pass

    # ----------------------------------------------------------------------
    @property
    @abstractmethod
    def bin_header(self):
        """"""
        pass

    # # ----------------------------------------------------------------------
    # @property
    # @abstractmethod
    # def sample_rate(self):
        # """"""
        # pass

    # ----------------------------------------------------------------------
    @property
    def eeg_buffer(self):
        """Return the deserialized data buffer."""

        return self._data_eeg

    # ----------------------------------------------------------------------
    @property
    def eeg_serial(self):
        """Return the binary data buffer."""

        return self._data_serial

    # ----------------------------------------------------------------------
    @property
    def eeg_pack(self):
        """Return the packed data buffer."""

        return self._data_packs

    # ----------------------------------------------------------------------
    def reset_buffers(self):
        """Discard buffers."""

        try:
            with self._data_eeg.mutex:
                self._data_eeg.queue.clear()
        except:
            while not self._data_eeg.empty():
                self._data_eeg.get()

        try:
            with self._data_serial.mutex:
                self._data_serial.queue.clear()
        except:
            self._data_serial.clear()
        else:
            while not self._data_serial.empty():
                self._data_serial.get()

        try:
            with self._data_packs.mutex:
                self._data_packs.queue.clear()
        except:
            while not self._data_packs.empty():
                self._data_packs.get()

    # ----------------------------------------------------------------------
    def start_collect(self, buffer_size=2**8, clear=True, raw=None):
        """Start a data collection asynchronously.

        Create a background process for read, deserialize and put data on a
        queue, the proces can be stoped colling `stop_collect`.

        Parameters
        ----------
        buffer_size: int, optional
            The size of characters for read from serial device.
        clear : bool, optional
            Reset buffer before start data collect.
        raw : bool, optional
            Ignore the deserialize process, all data is stored in binary format.
        """

        if clear:
            self.reset_buffers()

        if not raw is None:
            self.RAW = raw

        self.BOARDMODE = self.get_board_mode()
        self._daisy = self.daisy()
        self.READING = True
        self.start()  # start stream

        # Thread for read data
        if hasattr(self, "thread_data_collect") and self.thread_data_collect.isAlive():
            pass
        else:
            self.thread_data_collect = Thread(target=self.collect_data, args=(buffer_size, ))
            # self.thread_data_collect = Process(target=self.collect_data, args=(buffer_size, ))
            self.thread_data_collect.start()

        if not self.RAW:
            # Thread for deserialize data
            if hasattr(self, "thread_deserialize") and self.thread_deserialize.isAlive():
                pass
            else:
                self.thread_deserialize = Thread(target=self.deserialize_data)
                # self.thread_deserialize = Process(target=self.deserialize_data)
                self.thread_deserialize.start()

    # ----------------------------------------------------------------------
    def stop_collect(self):
        """Stop a data collection that run asynchronously."""

        self.READING = False
        self.stop()  # stop stream

    # ----------------------------------------------------------------------
    def collect(self, s=1, buffer_size=2**8, clear=True, raw=False):
        """Start a synchronous process for collect data.

        This call will hang until time completed.

        Parameters
        ----------
        s : int, optional
            Seconds for data collection.
        buffer_size: int, optional
            The size of characters for read from serial device.
        clear : bool, optional
            Reset buffer before start data collect.
        raw : bool, optional
            Ignore the deserialize process, all data is stored in binary format.
        """

        if clear:
            self.reset_buffers()

        self.RAW = raw

        self.start_collect(buffer_size)
        time.sleep(s)
        self.stop_collect()

    # ----------------------------------------------------------------------
    def collect_data(self, size=2**8):
        """Load binary data and put in a queue.

        For optimizations issues the data must be read in packages but write one
        to one in a queue, this method must be executed on a thread.

        Parameters
        ----------
        size : int, optional
            The buffer length for read.

        """
        # Flush input buffer, discarding all its contents.
        self.reset_input_buffer()

        # i = 0
        while self.READING:
            # i += 1
            # if i >= 100:
                # # self.all_is_right()
                # i = 0
            try:
                # [self._data_serial.append(data) for data in self.read(size)]
                self._data_serial.extend(self.read(size))
                # [self._data_serial.put(data) for data in self.read(size)]
            except SerialException as e:
                logging.error(e)

    # ----------------------------------------------------------------------
    def pack_data(self, milliseconds=100):
        """Create packages for specified time length.

        This process run in background.

        Parameters
        ----------
        milliseconds : int, optional
            The time length for packages.

        Returns
        -------
        int
            Number of samples for each package.

        """

        #samples = int(round(self.sample_rate * (milliseconds / 1000)))

        self.pack_time = milliseconds

        if hasattr(self, "thread_pack") and self.thread_pack.isAlive():
            pass
        else:
            self.thread_pack = Thread(target=self.pack_data_samples, args=(milliseconds, ))
            # self.thread_pack = Process(target=self.pack_data_samples, args=(milliseconds, ))
            self.thread_pack.start()

        # return samples

    # ----------------------------------------------------------------------
    def pack_data_samples(self, milliseconds):
        """Create packages for specified size.

        Parameters
        ----------
        samples : int, optional
            The number of samples for create packages.

        """

        while self.READING or not self._data_eeg.empty():

            pack = [self._data_eeg.get() for i in range(2)]
            while pack[-1][4] - pack[0][4] < timedelta(milliseconds=milliseconds):
                pack.extend([self._data_eeg.get() for i in range(SAMPLE_RATE_PACK)])

            pack_smp = np.array([p[0] for p in pack])
            pack_eeg = np.array([p[1] for p in pack])
            pack_aux = np.array([p[2] for p in pack])
            pack_foo = np.array([int(p[3], 16) for p in pack])
            pack_datetime = np.array([p[4].timestamp() for p in pack])
            # self._data_packs.put([pack_smp.T, pack_eeg, pack_aux.T, pack_foo.T, pack_datetime.T])
            self._data_packs.put([pack_smp, pack_eeg, pack_aux, pack_foo, pack_datetime])

        # Call again
        if self.READING:
            self.pack_data_samples(samples)

    # ----------------------------------------------------------------------
    def deserialize_data(self):
        """Extract data from binary format.

        The binary format is the more efficient way to get data, this method
        must be executed on a thread.

        """
        # # This deserialize call will end when there is no data to process or
        # # device is not in reading mode
        # # while self.READING and not self._data_serial.empty():
        # while self.READING:

            # while self._data_serial.qsize() < 100:
                # time.sleep(0.01)
                # pass

            # try:
                # # Wait for a start header
                # while self._data_serial.get(block=True, timeout=TIMEOUT_QUEUE) != self.bin_header:
                    # pass

                # while self._data_serial.qsize() < 50:
                    # time.sleep(0.01)
                    # pass

                # sample = self._data_serial.get(block=True, timeout=TIMEOUT_QUEUE)
                # eeg = [self._data_serial.get(block=True, timeout=TIMEOUT_QUEUE) for i in range(24)]
                # aux = [self._data_serial.get(block=True, timeout=TIMEOUT_QUEUE) for i in range(6)]
                # stop_byte = self._data_serial.get(block=True, timeout=TIMEOUT_QUEUE)
            # except Empty:
                # continue
            # try:
                # if hex(stop_byte)[-2].lower() != 'c':
                    # self._no_valid += 1
                    # logging.warning(f'{self}: Invalid stop byte {hex(stop_byte)} for sample {sample}, {100*(self._no_valid/self._valid):.3f}% loss')
                    # logging.warning(f'Data serial buffer: {self._data_serial.qsize()}')
                    # continue
                # else:
                    # self._valid += 1
                    # self.auto_drop()
            # except TypeError:
                # continue

        # This deserialize call will end when there is no data to process or
        # device is not in reading mode
        while self.READING:

            # # while len(self._data_serial) < 33:
                # # time.sleep(0.01)
                # # pass

            try:
                # Wait for a start header
                while self._data_serial.popleft() != self.bin_header:
                    pass

                # while len(self._data_serial) < 32:
                    # time.sleep(0.01)
                    # pass

                sample = self._data_serial.popleft()
                eeg = [self._data_serial.popleft() for i in range(24)]
                aux = [self._data_serial.popleft() for i in range(6)]
                stop_byte = self._data_serial.popleft()
            except:
                continue
            try:
                if hex(stop_byte)[-2].lower() != 'c':
                    self._no_valid += 1
                    # logging.warning(f'{self}: Invalid stop byte {hex(stop_byte)} for sample {sample}?, {100*(self._no_valid/self._valid):.3f}% loss')
                    # logging.warning(f'Data serial buffer: {len(self._data_serial)}')
                    continue
                else:
                    self._valid += 1
                    self.auto_drop(self._valid)
            except TypeError:
                continue

            # Scale factor to volts
            # first complete to 32-bit, Python not support 24-bit natively
            # finally scale with the correct gain for each channel
            eeg = sum([eeg[i:i + 3] + [0] for i in range(0, len(eeg), 3)], [])
            eeg = list(memoryview(bytearray(eeg)).cast('I'))  # I for 32-bit
            eeg = list(map(lambda _: _[0] * _[1],
                           zip(eeg, self.scale_factor_eeg)))  # scale

            # Update unsampled queue
            self._unsampled.get()
            self._unsampled.put(eeg)

            if self._daisy:
                m, n, o = list(self._unsampled.queue)
                # self._unsampled.get()

                if sample % 2:  # odd
                    eeg = np.average([m, o], axis=0).tolist() + n
                else:  # pair
                    eeg = n + np.average([m, o], axis=0).tolist()

            # Deserialize Aux Data
            aux = self.deserialize_aux(stop_byte, aux)

            # Return only channels selected
            eeg = [eeg[key] for key in self.montage.keys()]

            # The eeg buffer contain lists with data deserialized
            self._data_eeg.put([sample, eeg, aux, hex(stop_byte), datetime.now()])

            logging.debug(f'Sample {sample} ok')

        # Call again
        if self.READING:
            self.deserialize_data()

    # ---------------------------------------------------------------------
    def auto_drop(self, counter, samples=50):
        """Drop data automatically when they accumulate a lot."""
        return

        # if not (counter % samples):
            # std = len(self._data_serial)

            # if std > (33 * 90) * 5:
                # self._drop_proportion += 1# 0.1
            # elif std < (33 * 90) * 2:
                # self._drop_proportion -= 5# 0.5

            # for _ in range(33 * int(samples * self._drop_proportion)):
                # try:
                    # self._data_serial.popleft()
                # except:
                    # return

    # ----------------------------------------------------------------------
    def deserialize_aux(self, stop_byte, aux):
        """Determine the content of `AUX` bytes and format it.

        Auxialiar data could contain different kind of information: accelometer,
        user defined, time stamped and digital or analog inputs.
        The context of `AUX` bytes are determined by the stop byte.

        If `stop_byte` is `0xc0` the `AUX` bytes contain `Standard with accel`,
        this data are packaged at different frequency, they will be show up each
        10 or 11 packages, the final list will contain accelometer value in `G`
        units for axis `X`, `Y` and `Z` respectively and `None` when are not
        availables.

        If `stop_byte` is `0xc1` the `AUX` bytes contain `Standard with raw aux`,
        there are 3 types of raw data: `digital` in wich case the final list
        will contain the values for `D11`, `D12`, `D13`, `D17`, `D18`; `analog`
        with the values for `A7` (`D13`), `A6` (`D12`), `A5` (`D11`); `markers`
        data contain the the marker sended with `send_marker()` method.

        Parameters
        ----------
        stop_byte : int
             0xCX where X is 0-F in hex.

        aux : int
            6 bytes of data defined and parsed based on the `Footer` bytes.


        Returns
        -------
        list
            Correct data formated.

        """

        # Standard with accel
        if stop_byte == 0xc0:
            if aux.count(0) >= 3:
                return None
            else:
                return [0.002 * d / 16
                        for d in struct.unpack('>hhh', bytearray(aux))]

        # Standard with raw aux
        elif stop_byte == 0xc1:

            if self.BOARDMODE == b'analog':
                # A7, A6, A5
                # D13, D12, D11
                return aux[1::2]

            elif self.BOARDMODE == b'digital':
                # D11, D12, D13, D17, D18
                aux.pop(4)
                return aux

            elif self.BOARDMODE == b'marker':
                # Some time for some reason, marker not always send back from
                # OpenBCI, so this module implement a strategy to send a burst of
                # markers but read back only one.
                marker = aux.pop(1)
                if marker:
                    if (time.time() - self._last_marker) > 0.3:
                        self._last_marker = time.time()
                        return marker

                return 0

        # User defined
        elif stop_byte == 0xc2:
            pass

        # Time stamped set with accel
        elif stop_byte == 0xc3:
            pass

        # Time stamped with accel
        elif stop_byte == 0xc4:
            pass

        # Time stamped set with raw auxcalculate_sample_rate
        elif stop_byte == 0xc5:
            pass

        # Time stamped with raw aux
        elif stop_byte == 0xc6:
            pass

        return aux

    # ----------------------------------------------------------------------
    def calculate_sample_rate(self, seconds=5, buffer_size=2**8):
        """Calculate the sample rate with real data transmission.

        Parameters
        ----------
        seconds : int, optional
             Seconds for transmission.

        buffer_size : int, optional
            The size of characters for read from serial device.

        """

        # TODO: Calculate with timestamp

        # self.collect(seconds, buffer_size)
        # time.sleep(0.2)
        # sample_rate = self.eeg_buffer.qsize() / seconds
        # self.sample_rate_ = sample_rate



"""
Brython MDCComponent: MDCList
=============================


"""

from browser import html
from .core import MDCTemplate


########################################################################
class __listItem__(MDCTemplate):
    """"""

    MDC_optionals = {

        'meta': '<span class="mdc-list-item__meta">{meta}</span>',


        'icon_meta': '<a href="#" class="mdc-list-item__meta material-icons" onclick="event.preventDefault();" style="text-decoration: none; color: {meta_color};">{icon_meta}</a>',
        # 'fa_icon_meta': '<a href="#" class="mdc-list-item__meta material-icons" onclick="event.preventDefault();" style="text-decoration: none; color: {meta_color};">{icon_meta}</a>',


        'fa_icon_meta': '<i class="mdc-list-item__meta {fa_style_meta} {fa_icon_meta}" onclick="event.preventDefault();" style="text-decoration: none; color: {meta_color};"></i>',



        'icon': '<i class="material-icons mdc-list-item__graphic" aria-hidden="true">{icon}</i>',
        'fa_icon': '<i class="mdc-list-item__graphic {fa_style} {fa_icon}"></i>',



        'avatar': '<span class="mdc-list-item__graphic" style="background-color: {avatar_background_color}; color: {avatar_color};" role="presentation"><i class="material-icons" aria-hidden="true">{avatar}</i></span>',
        'placeholder': '<span class="mdc-list-item__graphic" style="background-color: {placeholder_background_color};"></span>',

    }

    # ----------------------------------------------------------------------

    def __new__(self, text, secondary_text=None, icon=False, icon_meta=False, meta=False, avatar=False, placeholder_background_coloiconr='rgba(0,0,0,.38)', avatar_color='white', meta_color='rgba(0,0,0,.38)', avatar_background_color='rgba(0,0,0,.38)', **kwargs):
        """"""

        if icon and icon.startswith('fa'):
            fa_style = icon[:icon.find('-')]
            fa_icon = 'fa' + icon[icon.find('-'):]
            del icon

        if icon_meta and icon_meta.startswith('fa'):
            fa_style_meta = icon_meta[:icon_meta.find('-')]
            fa_icon_meta = 'fa' + icon_meta[icon_meta.find('-'):]
            del icon_meta

        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""
        if context['secondary_text']:
            code = """
                <li class="mdc-list-item">
                  {icon}
                  {fa_icon}
                  {avatar}
                  {placeholder}
                  <span class="mdc-list-item__text">
                    <span class="mdc-list-item__primary-text">{text}</span>
                    <span class="mdc-list-item__secondary-text">{secondary_text}</span>
                  </span>
                  {meta}
                  {icon_meta}
                  {fa_icon_meta}
                </li>
            """

        else:
            code = """
                <li class="mdc-list-item">
                  {icon}
                  {avatar}
                  {placeholder}
                  <span class="mdc-list-item__text">{text}</span>
                  {meta}
                  {icon_meta}
                </li>
            """

        return cls.render_html(code, context)

    # ----------------------------------------------------------------------

    @classmethod
    def __getitem__(self, name):
        """"""
        if name is 'icon':
            return self.element.select('.mdc-list-item__graphic')[0]

        elif name is 'icon_meta':
            return self.element.select('.mdc-list-item__meta')[0]

        elif name is 'primary_text':
            return self.element.select('.mdc-list-item__primary-text')[0]


########################################################################
class __listChekItem__(MDCTemplate):
    """"""

    MDC_optionals = {

        'checked': 'checked=true',

    }

    # ----------------------------------------------------------------------

    def __new__(self, text, checked=False, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
            <li class="mdc-list-item checkbox-list-ripple-surface mdc-ripple-upgraded" style="--mdc-ripple-fg-size:360px; --mdc-ripple-fg-scale:1.6997692716423716; --mdc-ripple-fg-translate-start:258px, -163.06666564941406px; --mdc-ripple-fg-translate-end:120px, -156px;">
              <label for="trailing-checkbox-blueberries">{text}</label>
              <span class="mdc-list-item__meta">
                <div class="mdc-checkbox mdc-checkbox--upgraded mdc-ripple-upgraded mdc-ripple-upgraded--unbounded" style="--mdc-ripple-fg-size:24px; --mdc-ripple-fg-scale:1.6666666666666667; --mdc-ripple-left:8px; --mdc-ripple-top:8px;">
                  <input class="mdc-checkbox__native-control" {checked} id="trailing-checkbox-blueberries" type="checkbox">
                  <div class="mdc-checkbox__background">
                    <svg class="mdc-checkbox__checkmark" viewBox="0 0 24 24">
                      <path class="mdc-checkbox__checkmark-path" fill="none" stroke="white" d="M1.73,12.91 8.1,19.28 22.79,4.59"></path>
                    </svg>
                    <div class="mdc-checkbox__mixedmark"></div>
                  </div>
                </div>
              </span>
            </li>
        """
        return cls.render_html(code, context)


########################################################################
class MDCListGroup(MDCTemplate):
    """"""

    # ----------------------------------------------------------------------
    # def __new__(self, **kwargs):
        # """"""
        #self.element = self.render(locals(), kwargs)
        # return self.element

    # ----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
            <div class="mdc-list-group">
            </div>
        """
        return cls.render_html(code, context)

    # ----------------------------------------------------------------------

    @classmethod
    def add_list(cls, element, label, list_):
        """"""
        cls.element <= html.H3(label, Class='mdc-list-group__subheader')
        cls.element <= list_


########################################################################
class MDCList(MDCTemplate):
    """"""

    NAME = 'list', 'MDCList'

    MDC_optionals = {

        'two_line': 'mdc-list--two-line',
        'dense': 'mdc-list--two-line mdc-list--dense',
        'avatar': 'mdc-list--avatar-list',
        'non_interactive': 'mdc-list--non-interactive',

    }

    # ----------------------------------------------------------------------
    def __new__(self, two_line=False, dense=False, avatar=False, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
            <ul class="mdc-list {two_line} {dense} {avatar} {non_interactive}">
            </ul>
        """
        return cls.render_html(code, context)

    # ----------------------------------------------------------------------

    @classmethod
    def __getitem__(self, name):
        """"""
        # if name is 'actions':
            # return self.element.select('.mdc-card__actions')[0]

        # elif name is 'action_buttons':
            # return self.element.select('.mdc-card__action-buttons')[0]

        # elif name is 'action_icons':
            # return self.element.select('.mdc-card__action-icons')[0]

    # #----------------------------------------------------------------------
    # @classmethod
    # def add_action_button(cls, element, element, mdc, *args, **kwargs):
        # """"""

    # ----------------------------------------------------------------------
    @classmethod
    def add_item(cls, element, *args, **kwargs):
        """"""
        item = __listItem__(*args, **kwargs)
        cls.element <= item
        return item

    # ----------------------------------------------------------------------

    @classmethod
    def add_check_item(cls, element, *args, **kwargs):
        """"""
        item = __listChekItem__(*args, **kwargs)
        cls.element <= item
        return item

    # ----------------------------------------------------------------------

    @classmethod
    def add_divider(cls, element, hr=False, inset=False):
        """"""
        if inset:
            inset = 'mdc-list-divider--inset'
        else:
            inset = ''

        if hr:
            code = '<hr class="mdc-list-divider {inset}">'.format(inset=inset)
        else:
            code = '<li role="separator" class="mdc-list-divider {inset}"></li>'.format(inset=inset)
        code = cls.render_str(code)
        cls.element <= code




from browser import document
from radiant import WebSocket, get
import json

########################################################################


class OpenBCI(WebSocket):
    """"""

    # ----------------------------------------------------------------------
    def __init__(self, ip, ip_wifi, latency):
        """"""
        self.ip_wifi = ip_wifi
        self.latency = latency

        self.stream_id = document.select_one('#stream-id')
        # self.serial_buffer = document.select_one('#stream-serial_buffer')
        self.eeg_buffer = document.select_one('#stream-eeg_buffer')
        self.eeg_pack = document.select_one('#stream-eeg_pack')

        super().__init__(ip)

    # ----------------------------------------------------------------------
    def on_open(self, evt):
        """"""
        data = {"command": "set_up",
                'kwargs': {"device_id": self.ip_wifi,
                           "sample_rate": 'SAMPLE_RATE_250HZ',
                           }
                }

        self.send(data)

        data = {'command': 'start_stream',
                'kwargs': {'milliseconds': int(self.latency),
                           'boardmode': 'BOARD_MODE_MARKER',
                           }
                }

        self.send(data)

    # ----------------------------------------------------------------------
    def on_message(self, evt):
        """"""

        try:
            data = json.loads(evt.data[:300] + "]]}")
            # print(data.values())
            # print(data['pack_id'])
            self.stream_id.html = str(data['pack_id'])
            # self.serial_buffer.html = str(data['buffer_serial'])
            self.eeg_buffer.html = str(data['buffer_eeg'])
            self.eeg_pack.html = str(data['buffer_pack'])
        except:
            pass

        # print("Message received: {}".format(evt.data[:300]))

        # print(data['pack_id'])


########################################################################
class OpenBCI_API:
    """"""

    # ----------------------------------------------------------------------
    def __init__(self):
        """Constructor"""

        self.openbci = None

        self.button_connect = document.select_one("#btn-connect")
        self.button_connect.bind('click', self.connect)

        self.button_stop = document.select_one("#btn-stop")
        self.button_stop.bind('click', self.stop)

    # ----------------------------------------------------------------------
    def connect(self, evt):
        """"""
        self.ip_rasp = document.select_one('#input-ip_rasp')
        self.ip_wifi = document.select_one('#input-ip_wifi')
        self.latency = document.select_one('#input-latency')

        self.openbci = OpenBCI(f"ws://{self.ip_rasp.value}:8845/wifi", self.ip_wifi.value, self.latency.value)

        self.button_connect.style = {'display': 'none'}
        self.button_stop.style = {'display': 'block'}

        self.ip_rasp.attrs['disabled'] = True
        self.ip_wifi.attrs['disabled'] = True
        self.latency.attrs['disabled'] = True

    # ----------------------------------------------------------------------
    def stop(self, evt):
        """"""
        if self.openbci:
            self.openbci.close_connection()

        self.button_connect.style = {'display': 'block'}
        self.button_stop.style = {'display': 'none'}

        del self.ip_rasp.attrs['disabled']
        del self.ip_wifi.attrs['disabled']
        del self.latency.attrs['disabled']




import netifaces
import requests
import nmap


# ----------------------------------------------------------------------
def scan():
    """"""
    ip_list = {}
    try:
        local_wlan0 = netifaces.ifaddresses('wlan0')[netifaces.AF_INET][0]['addr']
    except:
        local_wlan0 = netifaces.ifaddresses('wlp2s0')[netifaces.AF_INET][0]['addr']

    nm = nmap.PortScanner()
    nm.scan(hosts=f'{local_wlan0}/24', arguments='-sn')
    hosts = nm.all_hosts()

    for host in hosts:
        try:
            response = requests.get(f'http://{host}/board', timeout=0.1)
            if response.ok:
                ip_list[host] = response.json
        except:
            continue

    return ip_list


if __name__ == '__main__':
    scan()
